##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

from Base import *
import CoreLvl2
TextReadTestCaseBase = CoreLvl2.TextReadTestCase
TextWriteTestCaseBase = CoreLvl2.TextWriteTestCase
NodeReadTestCaseBase = CoreLvl2.NodeReadTestCaseBase
NodeWriteTestCaseBase = CoreLvl2.NodeWriteTestCaseBase
del CoreLvl2

import xml.dom
from xml.dom import Node

# --- DocumentType

class DocumentTypeReadTestCase(NodeReadTestCaseBase):

    def setUp(self):
        self.doctype = self.node = self.implementation.createDocumentType(
            'foo:bar', 'uri:foo', 'uri:bar')

    def checkInternalSubset(self):
        # The DOM Level 2 recommendation is not clear on the value of the
        # internalSubset attribute when there isn't one; this test relies
        # on a clarification from Joe Kesselman:
        #
        # http://lists.w3.org/Archives/Public/www-dom/2001AprJun/0009.html
        #
        checkAttribute(self.doctype, 'internalSubset', None)
        checkReadOnly(self.doctype, 'internalSubset')

    def checkPublicId(self):
        checkAttribute(self.doctype, 'publicId', 'uri:foo')
        checkReadOnly(self.doctype, 'publicId')

    def checkSystemId(self):
        checkAttribute(self.doctype, 'systemId', 'uri:bar')
        checkReadOnly(self.doctype, 'systemId')

    def checkImportNode(self):
        foreignDoc = self.implementation.createDocument(None, 'foo', None)
        try:
            foreignDoc.importNode(self.doctype, 0)
        except xml.dom.NotSupportedErr:
            pass
        else:
            assert 0, "Was allowed to import a Document Type Node."


class DocumentTypeWriteTestCase(NodeWriteTestCaseBase):

    def setUp(self):
        self.doctype = self.node = self.implementation.createDocumentType(
            'foo:bar', 'uri:foo', 'uri:bar')


# --- ProcessingInstruction

class ProcessingInstructionReadTestCase(NodeReadTestCaseBase):

    def setUp(self):
        self.pi = self.createDocument().createProcessingInstruction("pit",
                                                                    "pid")
        self.node = self.pi

    def checkImportNode(self):
        foreignDoc = self.implementation.createDocument(None, 'foo', None)

        clone = foreignDoc.importNode(self.pi, 0)
        deepClone = foreignDoc.importNode(self.pi, 1)

        assert not isSameNode(self.pi, clone), "Clone is same as original."
        assert not isSameNode(self.pi, deepClone), "Clone is same as original."
        
        checkAttributeSameNode(clone, 'ownerDocument', foreignDoc)
        checkAttributeSameNode(deepClone, 'ownerDocument', foreignDoc)
        checkAttribute(clone, 'parentNode', None)
        checkAttribute(deepClone, 'parentNode', None)
        checkAttribute(clone, 'nodeType', self.pi.nodeType)
        checkAttribute(deepClone, 'nodeType', self.pi.nodeType)
        checkAttribute(clone, 'data', self.pi.data)
        checkAttribute(deepClone, 'data', self.pi.data)
        checkAttribute(clone, 'target', self.pi.target)
        checkAttribute(deepClone, 'target', self.pi.target)
        checkLength(clone.childNodes, 0)
        checkLength(deepClone.childNodes, 0)


class ProcessingInstructionWriteTestCase(NodeWriteTestCaseBase):

    def setUp(self):
        self.pi = self.createDocument().createProcessingInstruction("pit",
                                                                    "pid")
        self.node = self.pi


# --- CDATASection

class CDATASectionReadTestCase(TextReadTestCaseBase):

    def setUp(self):
        self.chardata = self.node = self.createDocument().createCDATASection(
            "com")


class CDATASectionWriteTestCase(TextWriteTestCaseBase):

    def setUp(self):
        self.chardata = self.node = self.createDocument().createCDATASection(
            "com")


# --- EntityReference

class EntityReferenceReadTestCase(NodeReadTestCaseBase):

    def setUp(self):
        self.entref = self.node = self.createDocument().createEntityReference(
            "eref")

    def checkImportNode(self):
        pass # TODO: Fill in meaningful test here.


class EntityReferenceWriteTestCase(NodeWriteTestCaseBase):

    def setUp(self):
        self.entref = self.node = self.createDocument().createEntityReference(
            "eref")


# --- Entity

class EntityReadTestCase(NodeReadTestCaseBase):

    def setUp(self):
        doc = self.document = self.parse("""
            <!DOCTYPE doc [
                <!ENTITY internalParsedE "Entity text">
            ]>
            <doc/>
        """)

        self.node = doc.doctype.entities.getNamedItem('internalParsedE')


class EntityWriteTestCase(NodeWriteTestCaseBase):

    def setUp(self):
        doc = self.document = self.parse("""
            <!DOCTYPE doc [
                <!ENTITY internalParsedE "Entity text">
            ]>
            <doc/>
        """)

        self.node = doc.doctype.entities.getNamedItem('internalParsedE')


# --- Notation

class NotationReadTestCase(NodeReadTestCaseBase):

    def setUp(self):
        doc = self.document = self.parse("""
            <!DOCTYPE doc [
                <!NOTATION aNotation PUBLIC "uri:public">
            ]>
            <doc/>
        """)

        self.node = doc.doctype.notations.getNamedItem('aNotation')


class NotationWriteTestCase(NodeWriteTestCaseBase):

    def setUp(self):
        doc = self.document = self.parse("""
            <!DOCTYPE doc [
                <!NOTATION aNotation PUBLIC "uri:public">
            ]>
            <doc/>
        """)

        self.node = doc.doctype.notations.getNamedItem('aNotation')


cases = buildCases(__name__, 'XML', '2.0')



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"Tests for garbage collection."
import os, sys
  
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing import ZopeTestCase
  
ZopeTestCase.installProduct('ParsedXML')

from Products.ParsedXML import ParsedXML, DOM

import App.ApplicationManager

class ReferenceTestCase(ZopeTestCase.ZopeTestCase):

    def afterSetUp(self):
        self.dbman = App.ApplicationManager.DebugManager() 

    def getRefcounts(self, ob):
        "return number of refcounts for instances of ob's class"
        name = '%s.%s' % (ob.__module__, ob.__class__.__name__)
        for i in self.dbman.refcount():
            if i[1] == name:
                return i[0]
        return 0

# XXX this leaks 1 refcount when I test it with python2.1
##     def testParsedXMLCollect(self):
##         "see if refcounts from ParsedXML product init are released"
##         doc = ParsedXML.ParsedXML('foo')
##         refcounts1 = self.getRefcounts(doc)
##         doc = ParsedXML.ParsedXML('foo')
##         refcounts2 = self.getRefcounts(doc)
##         assert refcounts2 == refcounts1, (
##             "ParsedXML leaked %d refcounts" % (refcounts2 - refcounts1))

    def testDOMParseCollect(self):
        "see if refcounts from DOM parse creation are released"
        testDir = os.path.join(
            sys.modules['Products.ParsedXML'].__path__[0],
            'tests')
        filename = os.path.join(testDir, 'xml', '4ohn4ktj.xml')
        doc = DOM.ExpatBuilder.parse(filename)
        refcounts1 = self.getRefcounts(doc)
        doc = DOM.ExpatBuilder.parse(filename)        
        refcounts2 = self.getRefcounts(doc)
        assert refcounts2 == refcounts1, (
            "DOM parse leaked %d refcounts" % (refcounts2 - refcounts1))

    def testDOMCreateCollect(self):
        "see if refcounts from DOM creation are released"
        doc = DOM.theDOMImplementation.createDocument(None, 'doc', None)
        refcounts1 = self.getRefcounts(doc)
        doc = DOM.theDOMImplementation.createDocument(None, 'doc', None)
        refcounts2 = self.getRefcounts(doc)
        assert refcounts2 == refcounts1, (
            "DOM parse leaked %d refcounts" % (refcounts2 - refcounts1))

if __name__ == '__main__':
    framework()
else:
    import unittest
    def test_suite():
        suite = unittest.TestSuite()
        suite.addTest(unittest.makeSuite(ReferenceTestCase))
        return suite



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

# Performance hacks:
#
#   -  Do not test node objects for truth; use an identity comparison
#      with None instead.  This avoids all attribute lookups and
#      requires exactly two dictionary lookups (module globals and
#      builtins).
#
#   -  There is no Node.__init__(); it doesn't do much for most
#      classes anyway.  This avoids a small performance penalty.

import string
_string = string
del string

import Acquisition
_Acquisition = Acquisition

def _reparent(ob, parent):
    ob.aq_inner.aq_parent = parent
    ob.aq_parent = parent
    if parent is None and ob._in_tree:
        ob.__dict__['_in_tree'] = 0

def _parent_of(ob, aq_inner=_Acquisition.aq_inner,
               aq_parent=_Acquisition.aq_parent):
    return aq_parent(aq_inner(ob))

_aq_base = _Acquisition.aq_base

del Acquisition

from ComputedAttribute import ComputedAttribute

import xml.dom


# legal qualified name pattern, from PyXML xml/dom/Document.py
# see http://www.w3.org/TR/REC-xml-names/#NT-QName
# we don't enforce namespace usage if using namespaces, which basically
# means that we don't disallow a leading ':'
# XXX there's more to the world than ascii a-z
# FIXME: should allow combining characters: fix when Python gets Unicode

try:
    import sre  # Note that Python 1.5.2's re module chokes.  Require sre.
except ImportError:
    def _ok_qualified_name(s, bad_first=(_string.digits + '.-'),
                           good=(_string.letters +
                                 _string.digits + '.-_:')):
        if len(s) < 1:
            return None
        if s[0] in bad_first:
            return None
        for c in s:
            if c not in good:
                return None
        return 1  # Indicates "passed".
else:
    _ok_qualified_name = sre.compile('[a-zA-Z_:][\w\.\-_:]*\Z', sre.UNICODE).match
    del sre

_TupleType = type(())

try:
    unicode
except NameError:
    _StringTypes = (type(''),)
else:
    # can't use u'' syntax due to Python 1.5.2 compatibility requirement
    _StringTypes = (type(''), type(unicode('')))


# http://www.w3.org/TR/1999/REC-xml-names-19990114/#ns-qualnames
def _check_qualified_name(name, uri='ok'):
    "test name for well-formedness"
    if _ok_qualified_name(name) is not None:
        if ":" in name:
            parts = _string.split(name, ":")
            if len(parts) != 2:
                raise xml.dom.NamespaceErr("malformed qualified name")
            if not (parts[0] and parts[1]):
                raise xml.dom.NamespaceErr("malformed qualified name")
            if not uri:
                raise xml.dom.NamespaceErr("no namespace URI for prefix")
        return 1
    else:
        raise xml.dom.InvalidCharacterErr()

# Common namespaces:
XML_NS = "http://www.w3.org/XML/1998/namespace"
XMLNS_NS = "http://www.w3.org/2000/xmlns/"

def _check_reserved_prefixes(prefix, namespaceURI):
    """
    Helper function to centralize the enforcement of reserved prefixes.
    Raises the appropriate NamespaceErr if the prefix is reserved but
    the namespaceURI doesn't match it.
    """
    if prefix == "xml" and namespaceURI != XML_NS:
        raise xml.dom.NamespaceErr(
            "illegal use of the 'xml' prefix")
    if prefix == "xmlns" and namespaceURI != XMLNS_NS:
        raise xml.dom.NamespaceErr(
            "illegal use of the 'xmlns' prefix")


# These are indexes into the list that is used to represent an Attr node.
_ATTR_NS = 0
_ATTR_NAME = 1
_ATTR_LOCALNAME = 2
_ATTR_PREFIX = 3
_ATTR_VALUE = 4
_ATTR_SPECIFIED = 5

# These are used for schema-derived information, and are not used for
# specified attributes.
_ATTR_TYPE = 6
_ATTR_REQUIRED = 7

_SUPPORTED_FEATURES = (
    ("core", None),
    ("xml", None),
    ("traversal", None),
    #("load", None),

    # According to DOM Erratum Core-14, the empty string should be
    # accepted as equivalent to null for hasFeature().
    ("core", ""),
    ("xml", ""),
    ("traversal", ""),
    #("load", ""),

    ("core", "1.0"),
    ("xml", "1.0"),

    ("core", "2.0"),
    ("xml", "2.0"),
    ("traversal", "2.0"),

    #("load", "3.0"),
    )


class _Dummy(_Acquisition.Explicit):
    pass


class DOMImplementation:
    def hasFeature(self, feature, version):
        feature = (_string.lower(feature), version)
        return feature in _SUPPORTED_FEATURES

    def createDocumentType(self, qualifiedName, publicId, systemId):
        _check_qualified_name(qualifiedName)
        import XMLExtended
        doctype = XMLExtended.DocumentType(qualifiedName, publicId, systemId)
        doctype = doctype.__of__(_Dummy())
        _reparent(doctype, None)
        return doctype

    def createDocument(self, namespaceURI, qualifiedName, docType=None):
        return Document(docType, namespaceURI, qualifiedName)

    # DOM Level 3 Core (working draft, 5 Jun 2001)

    def getAs(self, feature):
        return self

    # DOM Level 3 Load/Save (working draft, 9 Feb 2001)

    def createDOMBuilder(self):
        import LoadSave
        return LoadSave.DOMBuilder()

theDOMImplementation = DOMImplementation()


class AttributeControl:
    """Base class that provides reasonable get/set behavior for DOM
    classes."""

    _readonly = 0

    def __setattr__(self, name, value):
        setter = getattr(self, '_set_' + name, None)
        if setter is None:
            getter = getattr(self, '_get_' + name, None)
            if getter:
                raise xml.dom.NoModificationAllowedErr(
                    "read-only attribute: " + `name`)
            else:
                raise AttributeError, "no such attribute: " + `name`
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr(
                "cannot set attribute on read-only node")
        setter(value)


class Node(AttributeControl, _Acquisition.Explicit):
    ELEMENT_NODE = 1
    ATTRIBUTE_NODE = 2
    TEXT_NODE = 3
    CDATA_SECTION_NODE = 4
    ENTITY_REFERENCE_NODE = 5
    ENTITY_NODE = 6
    PROCESSING_INSTRUCTION_NODE = 7
    COMMENT_NODE = 8
    DOCUMENT_NODE = 9
    DOCUMENT_TYPE_NODE = 10
    DOCUMENT_FRAGMENT_NODE = 11
    NOTATION_NODE = 12

    # DOM Level 3 Core (working draft, 5 Jun 2001)
    # enum DocumentOrder
    DOCUMENT_ORDER_PRECEDING = 1
    DOCUMENT_ORDER_FOLLOWING = 2
    DOCUMENT_ORDER_SAME = 5
    DOCUMENT_ORDER_UNORDERED = 6

    # enum TreePosition
    TREE_POSITION_PRECEDING = 1
    TREE_POSITION_FOLLOWING = 2
    TREE_POSITION_ANCESTOR = 3
    TREE_POSITION_DESCENDANT = 4
    TREE_POSITION_SAME = 5
    TREE_POSITION_UNORDERED = 6

    attributes = None

    _children = ()
    _in_tree = 1
    _v_sibling_map = None

    namespaceURI = None
    prefix = localName = None
    nodeValue = None

    def _check_if_ancestor(self, node):
        "Helper function that raises if node is self or an ancestor of self."
        n = self
        node = _aq_base(node)        
        if _aq_base(n) is node:
            raise xml.dom.HierarchyRequestErr()
        if node._children:
            while n is not None:
                n = n.parentNode
                if _aq_base(n) is node:
                    raise xml.dom.HierarchyRequestErr()

    def __cmp__(self, other):
        # This doesn't seem to be getting used; a problem with
        # acquisition?  Jim thinks this may be an old bug that may
        # be resurfacing.  ;-(
        return cmp(id(_aq_base(self)), id(_aq_base(other)))

    def __hash__(self):
        return hash(id(_aq_base(self)))

    def __repr__(self):
        name = self.nodeName or "#"
        type = self.__class__.__name__
        if name[0] == "#":
            name = ""
        else:
            name = " " + `name`
        return "<%s%s at 0x%x>" % (
            type, name, id(_aq_base(self)))

    def _changed(self):
        # Mark the tree as changed this way since nodes can be
        # modified while they're not part of the tree (say, after
        # being removed but before being added somewhere else; the
        # real marking of the tree occurs when the node is re-inserted
        # into the tree).
        try:
            self.aq_acquire('__changed__')(1)
        except AttributeError:
            pass

    def _get_attributes(self):
        return

    def _get_childNodes(self):
        return ChildNodeList(self)
    childNodes = ComputedAttribute(_get_childNodes, 1)

    def _get_firstChild(self):
        if self._children:
            return self._children[0].__of__(self)
    firstChild = ComputedAttribute(_get_firstChild, 1)

    def _get_lastChild(self):
        if self._children:
            return self._children[-1].__of__(self)
    lastChild = ComputedAttribute(_get_lastChild, 1)

    def _get_localName(self):
        return self.localName

    def _get_namespaceURI(self):
        return self.namespaceURI

    def _get_nodeValue(self):
        return self.nodeValue

    def _get_nextSibling(self):
        node = self._getSiblingInfo()[1]
        if node is not None:
            return node.__of__(self.parentNode)
    nextSibling = ComputedAttribute(_get_nextSibling, 1)

    def _get_previousSibling(self):
        node = self._getSiblingInfo()[0]
        if node is not None:
            return node.__of__(self.parentNode)
    previousSibling = ComputedAttribute(_get_previousSibling, 1)

    def _getSiblingInfo(self):
        """
        Return a list containing the previous and next sibling of this node.
        If the sibling map doesn't exist on the parent yet, create it.
        """
        # Acquire the parent.
        if not self._in_tree:
            return [None, None]
        parent = _parent_of(self)
        if parent is None:
            return [None, None]

        # This implementation doesn't scale, but should amortize well
        # if .previousSibling and .nextSibling are actually used
        # much.  Is that enough?  This could be made lazier, but at
        # the expense of readability.
        #
        # The parent stores the sibling map for its children.
        sibmap = parent._v_sibling_map
        if sibmap is None:
            # There is no sibling map, create it.
            sibmap = {}
            parent.__dict__['_v_sibling_map'] = sibmap
        try:
            return sibmap[_aq_base(self)]
        except KeyError:
            # The sibling map hasn't been filled, fill it.
            # The *unwrapped* node is the key, the values are the
            # *wrapped* previous and next siblings.
            prev = None
            siblings = parent._children
            for i in range(len(siblings)):
                node = siblings[i]
                try:
                    next = siblings[i+1]
                except IndexError:
                    next = None
                sibmap[node] = [prev, next]
                prev = node
            return sibmap[_aq_base(self)]

    def _get_nodeName(self):
        return self.nodeName

    def _get_nodeType(self):
        return self.nodeType

    nodeValue = None

    def _get_ownerDocument(self):
        # We leverage Acquisition to get this from the
        # enclosing document.
        try:
            return self.aq_acquire('_acquireDocument')()
        except:
            return
    ownerDocument = ComputedAttribute(_get_ownerDocument, 1)

    def _get_parentNode(self):
        # Acquire the parent.
        if self._in_tree:
            return _parent_of(self)
    parentNode = ComputedAttribute(_get_parentNode, 1)

    def _check_prefix(self, value):
        "check prefix for wellformedness and validity"
        if ":" in value:
            raise xml.dom.NamespaceErr("':' not allowed in prefix")
        _check_qualified_name(value)
        if value is not None and not self.namespaceURI:
            raise xml.dom.NamespaceErr(
                "can't set prefix on a node without a namespace URI")
        if value == "xmlns":
            raise xml.dom.NamespaceErr(
                "can't use 'xmlns' as prefix")
        _check_reserved_prefixes(value, self.namespaceURI)
        if not self.namespaceURI:
            raise xml.dom.NamespaceErr(
                "no prefix allowed on nodes with no namespace URI")

    def _get_prefix(self):
        return self.prefix

    def _set_prefix(self, value):
        return

    def appendChild(self, newChild):
        if self._readonly or (
            newChild.parentNode and newChild.parentNode._readonly):
            raise xml.dom.NoModificationAllowedErr()
        if self.isSameNode(newChild):
            raise xml.dom.HierarchyRequestErr()
        # checking newChild._children here can avoid a lot of calls when
        # building a new tree
        if newChild._children:
            self._check_if_ancestor(newChild)

        children = self._children
        sibmap = self._v_sibling_map
        child = _aq_base(newChild)

        # Setup chidren if we don't have any, don't do this for Attr nodes.
        if not children and isinstance(children, _TupleType):
            self.__dict__['_children'] = children = []

        if child.nodeType == Node.DOCUMENT_FRAGMENT_NODE:
            if not child._children:
                return newChild
            for node in child._children:
                if node.nodeType not in self._allowed_child_types:
                    raise xml.dom.HierarchyRequestErr()
            for node in child._children:
                children.append(node)
            del child.__dict__['_children']
            if child.__dict__.has_key('_v_sibling_map'):
                del child.__dict__['_v_sibling_map']
            if self.__dict__.has_key('_v_sibling_map'):
                del self.__dict__['_v_sibling_map']
            self._changed()
            return newChild

        if child.nodeType not in self._allowed_child_types:
            raise xml.dom.HierarchyRequestErr()

        thisdoc = self.ownerDocument or self
        if not thisdoc.isSameNode(newChild.ownerDocument):
            raise xml.dom.WrongDocumentErr()

        # Save the unwrapped child
        children.append(child)

        # Reparent the child
        parent = newChild.parentNode
        if parent is not None:
            parent.removeChild(newChild)
        _reparent(newChild, self)
        if child.__dict__.has_key('_in_tree'):
            del child.__dict__['_in_tree']

        # delete the sibling map, it will be recreated next use        
        if self.__dict__.has_key('_v_sibling_map'):
            del self.__dict__['_v_sibling_map']

        # Notify our containing database object that we have changed
        self._changed()

        return newChild

    def insertBefore(self, newChild, refChild):
        if refChild is None:
            return self.appendChild(newChild)
        if self._readonly or (
            newChild.parentNode and newChild.parentNode._readonly):
            raise xml.dom.NoModificationAllowedErr()
        if self.isSameNode(newChild):
            raise xml.dom.HierarchyRequestErr()
        if self.nodeType == Node.DOCUMENT_NODE:
            if (newChild.ownerDocument is not None
                and not self.isSameNode(newChild.ownerDocument)):
                raise xml.dom.WrongDocumentErr()
        else:
            thisdoc = self.ownerDocument
            thatdoc = newChild.ownerDocument
            if thatdoc is None:
                if thisdoc.isSameNode(newChild):
                    raise xml.dom.HierarchyRequestErr()
            elif not thisdoc.isSameNode(thatdoc):
                raise xml.dom.WrongDocumentErr()
        self._check_if_ancestor(newChild)
        if newChild.nodeType == Node.DOCUMENT_FRAGMENT_NODE:
            # This is destructive of the fragment, but I think that's ok.
            # Note that the call to tuple() is required, or a more tedious
            # loop construct would have to be used.
            for child in tuple(newChild._children):
                self.insertBefore(child.__of__(newChild), refChild)
            return newChild
        if newChild.nodeType not in self._allowed_child_types:
            raise xml.dom.HierarchyRequestErr()
        children = self._children
        # setup children if we don't have any; don't do this for Attr nodes
        if not children and isinstance(children, _TupleType):
            self.__dict__['_children'] = children = []
        if newChild.isSameNode(refChild):
            return newChild
        if newChild.parentNode:
            newChild.parentNode.removeChild(newChild)
        
        ref = _aq_base(refChild)
        try:
            i = children.index(ref)
        except ValueError:
            raise xml.dom.NotFoundErr()

        _reparent(newChild, self)
        new = _aq_base(newChild)
        if new.__dict__.has_key('_in_tree'):
            del new.__dict__['_in_tree']
        children.insert(i, new)
        self._changed()
        sibmap = self._v_sibling_map
        if sibmap:
            del self.__dict__['_v_sibling_map']
            #prev, next = sibmap[ref]
            #if prev:
            #    sibmap[prev][1] = new
            #sibmap[new] = [prev, ref]
            #sibmap[ref][0] = new
        return newChild

    def removeChild(self, oldChild):
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        child = _aq_base(oldChild)
        if self._children:
            children = self._children
            try:
                i = children.index(child)
            except ValueError:
                raise xml.dom.NotFoundErr()
            del children[i]
            sibmap = self._v_sibling_map
            if sibmap:
                del self.__dict__['_v_sibling_map']
                #prev, next = sibmap[child]
                #if prev:
                #    sibmap[prev][1] = next
                #if next:
                #    sibmap[next][0] = prev
                #del sibmap[child]
            self._changed()
        else:
            raise xml.dom.NotFoundErr()

        child.__dict__['_in_tree'] = 0
        return oldChild

    def replaceChild(self, newChild, oldChild):
        if self._readonly or (
            newChild.parentNode and newChild.parentNode._readonly):
            raise xml.dom.NoModificationAllowedErr()
        if self.isSameNode(newChild):
            raise xml.dom.HierarchyRequestErr()
        if self.nodeType == Node.DOCUMENT_NODE:
            if newChild.ownerDocument \
               and not self.isSameNode(newChild.ownerDocument):
                raise xml.dom.WrongDocumentErr()
        elif self.ownerDocument and newChild.ownerDocument \
           and not self.ownerDocument.isSameNode(newChild.ownerDocument):
            raise xml.dom.WrongDocumentErr()
        # Check for HierarchyRequestErr here so that we can fail
        # before mutating the currrent node:
        if newChild.nodeType == Node.DOCUMENT_FRAGMENT_NODE:
            children = newChild._children
            for child in children:
                if child.nodeType not in self._allowed_child_types:
                    raise xml.dom.HierarchyRequestErr()
        elif newChild.nodeType not in self._allowed_child_types:
            raise xml.dom.HierarchyRequestErr()
        self._check_if_ancestor(newChild)
        next = oldChild.nextSibling
        self.removeChild(oldChild)
        if next:
            if newChild.nodeType == Node.DOCUMENT_FRAGMENT_NODE:
                nodes = newChild._children
                while nodes:
                    node = nodes[0].__of__(newChild.parentNode)
                    self.insertBefore(node, next)
            else:
                self.insertBefore(newChild, next)
        else:
            if newChild.nodeType == Node.DOCUMENT_FRAGMENT_NODE:
                nodes = newChild._children
                while nodes:
                    node = nodes[0].__of__(newChild.parentNode)
                    self.appendChild(node)
            else:
                self.appendChild(newChild)
        return oldChild

    def _mergeChildList(self, children):
        """helper function for normalize.  Merge the text nodes in children
        in place according to normalize, normalizing element children as well,
        but doesn't do all of the child list housekeeping.  Returns true
        if changes were made.
        Works for element child lists and attr child lists."""
        changed = 0
        i = 0
        L = []
        for child in children:
            if child.nodeType == Node.TEXT_NODE:
                if child.data == "":
                    # drop this child
                    child.__dict__['_in_tree'] = 0
                    changed = 1
                elif L and L[-1].nodeType == child.nodeType:
                    # merge this child with previous sibling
                    data = L[-1].data + child.data
                    d = L[-1].__dict__
                    d['data'] = d['nodeValue'] = data
                    child.__dict__['_in_tree'] = 0
                    changed = 1
                else:
                    L.append(child)
            elif (child.nodeType == Node.ELEMENT_NODE
                  and child._children):
                child.normalize()
                L.append(child)
            else:
                L.append(child)
        if changed:
            children[:] = L
        return changed

    def normalize(self):
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        aChanged = 0
        d = self.__dict__
        if d.has_key('_attributes'):
            attributes = self._attributes
            for attr in attributes:
                attrVal = attributes[0][_ATTR_VALUE]
                if type(attrVal) in _StringTypes:
                    continue
                if len(attrVal) == 1:
                    child = attrVal[0]
                    if child.nodeType == Node.TEXT_NODE and not child.data:
                        del child
                        aChanged = 1
                else:
                    aChanged = self._mergeChildList(attrVal) or aChanged
        children = self._children
        if not children:
            return
        if len(children) == 1:
            child = children[0]
            if child.nodeType == Node.TEXT_NODE and not child.data:
                self.removeChild(child.__of__(self))
            elif (child.nodeType == Node.ELEMENT_NODE
                  and child._children):
                child.__of__(self).normalize()
            return
        cChanged = self._mergeChildList(children)
        if cChanged:
            if d.has_key('_v_sibling_map'):
                # this is now invalid; let it be recreated on demand
                del d['_v_sibling_map']
        if cChanged or aChanged:
            self._changed()

    def hasAttributes(self):
        return 0

    def hasChildNodes(self):
        return self._children and 1 or 0

    def isSupported(self, feature, version):
        if self.ownerDocument:
            impl = self.ownerDocument.implementation
        else:
            impl = theDOMImplementation
        return impl.hasFeature(feature, version)

    def cloneNode(self, deep):
        node = _aq_base(self)
        clone = node._cloneNode(deep and 1 or 0,
                                mutable=1, document=self.ownerDocument)
        clone.__dict__['_in_tree'] = 0
        clone = clone.__of__(self)
        return clone

    def _cloneNode(self, deep, mutable, document):
        # self is *not* an acquisition wrapper!
        clone = self.__class__.__basicnew__()
        d = clone.__dict__
        d.update(self.__dict__)
        if deep:
            if self._children:
                # make a recursive clone:
                d['_children'] = L = []
                for child in self._children:
                    L.append(child._cloneNode(deep, mutable, document))
        elif d.has_key('_children'):
            del d['_children']
        if d.has_key('_v_sibling_map'):
            del d['_v_sibling_map']
        if mutable and d.has_key('_readonly'):
            del d['_readonly']
        return clone

    # DOM Level 3 (Working Draft, 5 Jun 2001)

    def _get_baseURI(self):
        node = self
        d = node.__dict__
        while not d.has_key('baseURI'):
            node = self.parentNode
            if node is None:
                return
            d = node.__dict__
        return d['baseURI']
    baseURI = ComputedAttribute(_get_baseURI, 1)

    def getAs(self, feature):
        return self

    def isSameNode(self, other):
        # This is useful since cmp() (hence ==, !=) don't seem to work
        # with acquisition.
        return (_aq_base(self) is _aq_base(other))

    def lookupNamespacePrefix(self, namespaceURI):
        node = self
        while node is not None:
            ns_map = getattr(self, '_ns_uri_prefixes', None)
            if ns_map and ns_map.has_key(namespaceURI):
                return ns_map[namespaceURI][0]
            node = node.parentNode

    def lookupNamespaceURI(self, prefix):
        node = self
        while node is not None:
            ns_map = getattr(self, '_ns_prefix_uri', None)
            if ns_map and ns_map.has_key(prefix):
                return ns_map[prefix]
            node = node.parentNode


class Parentless:
    """Node mixin that doesn't have a parent node."""

    parentNode = None
    nextSibling = None
    previousSibling = None

    def _get_parentNode(self):
        return

    def _get_previousSibling(self):
        return

    def _get_nextSibling(self):
        return

    def _getSiblingInfo(self):
        return [None, None]


class Childless:
    """Node mixin that doesn't allow child nodes.

    This ensures safety when used as a base class for node types that
    should never have children of their own, and allows slightly
    faster response for some methods.
    """

    _allowed_child_types = ()

    def _get_firstChild(self):
        return
    firstChild = None

    def _get_lastChild(self):
        return
    lastChild = None

    def appendChild(self, newChild):
        raise xml.dom.HierarchyRequestErr()

    def insertBefore(self, newChild, oldChild):
        raise xml.dom.HierarchyRequestErr()

    def removeChild(self, oldChild):
        raise xml.dom.NotFoundErr()

    def replaceChild(self, newChild, oldChild):
        # This could reasonably raise NotFoundErr as well.
        raise xml.dom.HierarchyRequestErr()

    def hasChildNodes(self):
        return 0

    def normalize(self):
        return


class TextualContent:
    """Mixin class defining the recursive support for textContent
    needed for some types of container nodes.
    """
    # DOM Level 3 (working draft, 5 June 2001)

    def _get_textContent(self):
        L = []
        for node in self.childNodes:
            nodeType = node.nodeType
            if (  nodeType == Node.COMMENT_NODE
                  or nodeType == Node.PROCESSING_INSTRUCTION_NODE):
                continue
            if nodeType == Node.TEXT_NODE:
                text = node.data
            else:
                text = node._get_textContent()
            L.append(text)
        if L:
            return _string.join(L, '')
        else:
            return ''
    textContent = ComputedAttribute(_get_textContent, 1)


class Document(Parentless, TextualContent, Node):
    nodeName = "#document"
    nodeType = Node.DOCUMENT_NODE

    _allowed_child_types = (Node.ELEMENT_NODE,
                            Node.COMMENT_NODE,
                            Node.PROCESSING_INSTRUCTION_NODE)

    _doctype = None
    _in_tree = 0
    implementation = theDOMImplementation

    def __init__(self, doctype, namespaceURI, qualifiedName):
        _check_qualified_name(qualifiedName, namespaceURI)
        if namespaceURI:
            e = Element(namespaceURI, qualifiedName, element_id=0)
        else:
            e = Element(None, qualifiedName, element_id=0)
        del e.__dict__['_in_tree']
        if doctype:
            if doctype._in_tree:
                raise xml.dom.WrongDocumentErr()
            doctype = _aq_base(doctype)
            self.__dict__['_doctype'] = doctype
            del doctype.__dict__['_in_tree']
            L = [doctype, e]
        else:
            L = [e]
        self.__dict__['_children'] = L
        self.__dict__['_attr_info'] = {}
        self.__dict__['_element_count'] = 1
        
    def _set_nodeValue(self, data):
        return None

    def _get_doctype(self):
        if self._doctype is None:
            return
        else:
            return self._doctype.__of__(self)
    doctype = ComputedAttribute(_get_doctype)

    def _get_implementation(self):
        return self.implementation

    def _get_documentElement(self):
        for node in self._children:
            if node.nodeType == Node.ELEMENT_NODE:
                return node.__of__(self)
    documentElement = ComputedAttribute(_get_documentElement)

    ownerDocument = None
    def _get_ownerDocument(self):
        return

    def _acquireDocument(self):
        # This method gets acquired by descendents that want their
        # owner; the result is an unwrapped Document node.
        return self

    # child insertion methods:
    # check for adding 2nd element node; other checks are elsewhere

    # helper method for 2nd element node detection
    def _hasDocumentElement(self):
        for node in self._children:
            if node.nodeType == Node.ELEMENT_NODE:
                return 1

    def appendChild(self, newChild):
        if newChild.nodeType == Node.DOCUMENT_TYPE_NODE:
            raise xml.dom.HierarchyRequestErr(
                "cannot change document type via tree manipulation")
        if (newChild.nodeType == Node.ELEMENT_NODE and self._children
            and self._hasDocumentElement()):
            raise xml.dom.HierarchyRequestErr()
        return Node.appendChild(self, newChild)

    def insertBefore(self, newChild, refChild):
        if newChild.nodeType == Node.DOCUMENT_TYPE_NODE:
            raise xml.dom.HierarchyRequestErr(
                "cannot change document type via tree manipulation")
        if (newChild.nodeType == Node.ELEMENT_NODE and self._children
            and self._hasDocumentElement()):
            raise xml.dom.HierarchyRequestErr()
        return Node.insertBefore(self, newChild, refChild)

    def removeChild(self, oldChild):
        if oldChild.nodeType == Node.DOCUMENT_TYPE_NODE:
            raise xml.dom.NoModificationAllowedErr(
                "cannot change document type via tree manipulation")
        return Node.removeChild(self, oldChild)

    def replaceChild(self, newChild, oldChild):
        if (newChild.nodeType == Node.ELEMENT_NODE and self._children
            and oldChild.nodeType != Node.ELEMENT_NODE
            and self._hasDocumentElement()):
            raise xml.dom.HierarchyRequestErr()
        if (newChild.nodeType == Node.DOCUMENT_TYPE_NODE
            or oldChild.nodeType == Node.DOCUMENT_TYPE_NODE):
            raise xml.dom.HierarchyRequestErr(
                "cannot change document type via tree manipulation")
        return Node.replaceChild(self, newChild, oldChild)        

    # unknown acquisition environment, no computedAttribute unwrapping
    # so we must redefine these methods
    def _get_childNodes(self):
        return ChildNodeList(self)
    childNodes = ComputedAttribute(_get_childNodes)

    def _get_firstChild(self):
        if self._children:
            return self._children[0].__of__(self)
    firstChild = ComputedAttribute(_get_firstChild)

    def _get_lastChild(self):
        if self._children:
            return self._children[-1].__of__(self)
    lastChild = ComputedAttribute(_get_lastChild)

    def createAttribute(self, name):
        _check_qualified_name(name)
        return Attr([None, name, None, None, "", 1], self).__of__(self)

    def createAttributeNS(self, namespaceURI, qualifiedName):
        _check_qualified_name(qualifiedName, namespaceURI)
        names = _string.split(qualifiedName, ":", 1)
        localName = names[-1]
        if len(names) > 1:
            prefix = names[0]
            _check_reserved_prefixes(prefix, namespaceURI)
        else:
            prefix = None
        item = [namespaceURI, qualifiedName, localName, prefix, "", 1]
        return Attr(item, self).__of__(self)

    def createCDATASection(self, data):
        import XMLExtended
        return XMLExtended.CDATASection(data).__of__(self)

    def createComment(self, data):
        return Comment(data).__of__(self)

    def createDocumentFragment(self):
        return DocumentFragment(self).__of__(self)

    def createElement(self, tagName):
        _check_qualified_name(tagName)
        e = Element(None, tagName, element_id=self._element_count).__of__(self)
        self.__dict__['_element_count'] = self._element_count + 1
        if self._attr_info.has_key(tagName):
            e.__dict__['_attr_info'] = self._attr_info[tagName]
        self._changed()
        return e

    def createElementNS(self, namespaceURI, qualifiedName, stuff=None):
        _check_qualified_name(qualifiedName, namespaceURI)
        if ":" in qualifiedName and not namespaceURI:
            raise xml.dom.NamespaceErr(
                "tag name with prefix has no namespace URI")
        e = Element(namespaceURI, qualifiedName, stuff,
                    element_id=self._element_count).__of__(self)
        self.__dict__['_element_count'] = self._element_count + 1
        if self._attr_info.has_key(qualifiedName):
            e.__dict__['_attr_info'] = self._attr_info[qualifiedName]
        self._changed()
        return e

    def createEntityReference(self, name):
        if not _ok_qualified_name(name):
            raise xml.dom.InvalidCharacterErr()
        import XMLExtended
        return XMLExtended.EntityReference(name).__of__(self)

    def createProcessingInstruction(self, target, data):
        if not _ok_qualified_name(target):
            raise xml.dom.InvalidCharacterErr()
        if _string.lower(target) == "xml":
            raise xml.dom.InvalidCharacterErr(
                "'%s' not allowed as a processing instruction target"
                % target)
        import XMLExtended
        return XMLExtended.ProcessingInstruction(target, data).__of__(self)

    def createTextNode(self, data):
        return Text(data).__of__(self)

    def getElementById(self, elementId):
        # This performs a depth-first search of the tree for every request
        # (if any ID attributes are defined); this is necessary in order
        # to create the proper chain of acquisition wrappers.
        #
        info = self._compute_id_map()
        if not info:
            return
        queue = [self.documentElement]
        while queue:
            elem = queue.pop(0)
            if info.has_key(elem.tagName):
                attrs = info[elem.tagName]
                for name in attrs:
                    if elem.getAttribute(name) == elementId:
                        return elem
            if elem.hasChildNodes():
                childNodes = elem.childNodes
                L = []
                for node in childNodes:
                    if node.nodeType == Node.ELEMENT_NODE:
                        L.append(node)
                queue[:0] = L

    def _compute_id_map(self):
        # Returns a mapping from tagName to a list of attribute names
        # that bear IDs.  The computed value is cached on the instance;
        # if this is used during a transaction that modifies the document
        # the structure will be saved as an "accidental" side effect.
        # Since the DOM does not support changing the content model, this
        # is acceptable.
        try:
            return self._id_info
        except AttributeError:
            self.__dict__['_id_info'] = info = {}
            for tagName, L in self._attr_info.items():
                for item in L:
                    if item[_ATTR_TYPE] == "ID":
                        if info.has_key(tagName):
                            info[tagName].append(item[_ATTR_NAME])
                        else:
                            info[tagName] = [item[_ATTR_NAME]]
            return info

    def getElementsByTagName(self, tagName):
        nodeList = SimpleNodeList()
        _getElementsByTagNameHelper(self, tagName, nodeList._data)
        return nodeList

    def getElementsByTagNameNS(self, namespaceURI, localName):
        nodeList = SimpleNodeList()
        _getElementsByTagNameNSHelper(
            self, namespaceURI, localName, nodeList._data)
        return nodeList

    def isSupported(self, feature, version):
        return self.implementation.hasFeature(feature, version)

    def importNode(self, importedNode, deep):
        if importedNode.nodeType in (
            Node.DOCUMENT_NODE, Node.DOCUMENT_TYPE_NODE):
            raise xml.dom.NotSupportedErr(
                "can't import this kind of node")
        doc = importedNode.ownerDocument
        if doc.implementation == self.implementation:
            # same implementation, so we're in good shape
            node = _aq_base(importedNode)
            clone = node._cloneNode(deep and 1 or 0, mutable=1, document=self)
            clone.__dict__['_in_tree'] = 0
            clone = clone.__of__(self)
            if hasattr(clone, '_set_owner_document'):
                clone._set_owner_document(self)
            return clone
        raise xml.dom.NotSupportedErr(
            "can't import from a different DOM implementation")

    # DOM Level 2 Traversal

    def createNodeIterator(self, root, whatToShow, filter,
                           entityReferenceExpansion):
        import Traversal
        return Traversal.NodeIterator(root, whatToShow, filter,
                                      entityReferenceExpansion)

    def createTreeWalker(self, root, whatToShow, filter,
                         entityReferenceExpansion):
        import Traversal
        return Traversal.TreeWalker(root, whatToShow, filter,
                                    entityReferenceExpansion)

    # DOM Level 3 (Working Draft, 5 Jun 2001)
    # I expect some or all of these will become read-only before the
    # recommendation is finished.
    actualEncoding = None
    encoding = None
    standalone = 0
    strictErrorChecking = 0
    version = None

    # Override the inherited handler for textContent since the
    # acquisition context is different.
    textContent = ComputedAttribute(TextualContent._get_textContent)

    def _get_actualEncoding(self):
        return self.actualEncoding
    def _set_actualEncoding(self, value):
        self.__dict__['actualEncoding'] = value
        self._changed()

    def _get_encoding(self):
        return self.encoding
    def _set_encoding(self, value):
        self.__dict__['encoding'] = value
        self._changed()

    def _get_standalone(self):
        return self.standalone
    def _set_standalone(self, value):
        self.__dict__['standalone'] = value and 1 or 0
        self._changed()

    def _get_strictErrorChecking(self):
        return self.strictErrorChecking
    def _set_strictErrorChecking(self, value):
        self.__dict__['strictErrorChecking'] = value and 1 or 0
        self._changed()

    def _get_version(self):
        return self.version
    def _set_version(self, value):
        self.__dict__['version'] = value
        self._changed()

    def normalizeNS(self):
        pass

    def setBaseURI(self, baseURI):
        # we really need something like urlparse.isabs()!
        if ':' not in baseURI:
            raise xml.dom.SyntaxErr("baseURI is not an absolute URI")
        self.__dict__['baseURI'] = baseURI
        self._changed()


def _getElementsByTagNameHelper(parent, name, list):
    for node in parent._children:
        if node.nodeType == Node.ELEMENT_NODE:
            if (name == "*" or node.tagName == name):
                list.append(node.__of__(parent))
            _getElementsByTagNameHelper(node.__of__(parent), name, list)

def _getElementsByTagNameNSHelper(parent, nsURI, localName, list):
    for node in parent._children:
        if node.nodeType == Node.ELEMENT_NODE:
            if ((localName == "*" or node.localName == localName) and
                (nsURI == "*" or node.namespaceURI == nsURI)):
                list.append(node.__of__(parent))
            _getElementsByTagNameNSHelper(node.__of__(parent),
                                          nsURI, localName, list)


class DocumentFragment(Parentless, TextualContent, Node):
    nodeName = "#document-fragment"
    nodeType = Node.DOCUMENT_FRAGMENT_NODE
    parentNode = None

    _in_tree = 0

    _allowed_child_types = (Node.ELEMENT_NODE,
                            Node.TEXT_NODE,
                            Node.PROCESSING_INSTRUCTION_NODE,
                            Node.COMMENT_NODE,
                            Node.CDATA_SECTION_NODE,
                            Node.ENTITY_REFERENCE_NODE)

    def __init__(self, owner):
        self.__dict__['ownerDocument'] = owner

    def _set_nodeValue(self, data):
        return None

    def _get_ownerDocument(self):
        return self.ownerDocument

    def _get_parentNode(self):
        return

    def _set_owner_document(self, doc):
        self.__dict__['ownerDocument'] = doc


def _split_qname(namespaceURI, qualifiedName):
    if ":" in qualifiedName:
        prefix, localName = _string.split(qualifiedName, ':', 1)
        if prefix == "xml" and namespaceURI != XML_NS:
            raise xml.dom.NamespaceErr(
                "illegal use of the 'xml' prefix")
        if prefix == "xmlns" and namespaceURI != XMLNS_NS:
            raise xml.dom.NamespaceErr(
                "illegal use of the 'xmlns' prefix")
        return prefix, localName
    else:
        return None, qualifiedName


# Element _attribute members can be shared with Attr nodes; see comment
# at the Attr class.
class Element(TextualContent, Node):
    nodeType = Node.ELEMENT_NODE

    _allowed_child_types = (Node.ELEMENT_NODE,
                            Node.TEXT_NODE,
                            Node.COMMENT_NODE,
                            Node.PROCESSING_INSTRUCTION_NODE,
                            Node.CDATA_SECTION_NODE,
                            Node.ENTITY_REFERENCE_NODE)

    _attributes = ()
    _attr_info = ()

    def __init__(self, namespaceURI, qualifiedName, stuff=None, element_id=0):
        d = self.__dict__
        d['_in_tree'] = 0
        d['_element_id'] = element_id
        d['nodeName'] = qualifiedName
        d['tagName'] = qualifiedName
        if stuff:
            d['namespaceURI'] = namespaceURI
            d['prefix'] = stuff[0]
            d['localName'] = stuff[1]
        elif namespaceURI:
            d['namespaceURI'] = namespaceURI
            prefix, localName = _split_qname(namespaceURI, qualifiedName)
            d['prefix'] = prefix
            d['localName'] = localName

    def _cloneNode(self, deep, mutable, document):
        clone = Node._cloneNode(self, deep, mutable, document)
        d = clone.__dict__
        # create element_id for cloned element
        d['_element_id'] = document._element_count
        document.__dict__['_element_count'] = document._element_count + 1
        # we need to notify changes as we change element_count
        self._changed()
        
        info = document._attr_info.get(self.tagName)
        if info:
            d['_attr_info'] = info
        elif clone._attr_info:
            del d['_attr_info']
            
        if d.has_key('_attributes'):
            d['_attributes'] = attrs = map(list, self._attributes)
            for i in range(len(attrs) - 1, -1, -1):
                item = attrs[i]
                if not item[_ATTR_SPECIFIED]:
                    del attrs[i]
        return clone

    def _set_nodeValue(self, data):
        return None

    def _set_prefix(self, value):
        self._check_prefix(value)
        d = self.__dict__
        d['prefix'] = value
        s = "%s:%s" % (value, self.localName)
        d['nodeName'] = s
        # Do this here to avoid paying for a method call
        d['tagName'] = s

    def _get_tagName(self):
        return self.tagName

    tagName = ComputedAttribute(_get_tagName, 1)

    # not DOM, but convenience elementId accessor. Each element node
    # has a document-unique element_id
    def _get_elementId(self):
        return self._element_id

    elementId = ComputedAttribute(_get_elementId, 1)
    
    def getElementsByTagName(self, tagName):
        nodeList = SimpleNodeList()
        _getElementsByTagNameHelper(self, tagName, nodeList._data)
        return nodeList

    def getElementsByTagNameNS(self, namespaceURI, localName):
        nodeList = SimpleNodeList()
        _getElementsByTagNameNSHelper(
            self, namespaceURI, localName, nodeList._data)
        return nodeList

    def _get_attributes(self):
        return AttributeMap(self)
    attributes = ComputedAttribute(_get_attributes, 1)

    def hasAttributes(self):
        return ((self._attributes or self._attr_info) and 1 or 0)

    def getAttribute(self, name):
        for item in self._attributes:
            if name == item[_ATTR_NAME]:
                if type(item[_ATTR_VALUE]) in _StringTypes:
                    # value is a string
                    return item[_ATTR_VALUE]
                # value is a subtree
                return _attr_get_value(item[_ATTR_VALUE])
        for item in self._attr_info:
            if name == item[_ATTR_NAME]:
                # will always be a string in the current implementation
                return item[_ATTR_VALUE]
        return ""

    def getAttributeNode(self, name):
        for item in self._attributes:
            if name == item[_ATTR_NAME]:
                doc = self.ownerDocument
                return Attr(item, doc, self).__of__(self)
        for item in self._attr_info:
            if name == item[_ATTR_NAME]:
                item = item[:]
                if self._attributes:
                    self._attributes.append(item)
                else:
                    self.__dict__['_attributes'] = [item]
                return Attr(item, self.ownerDocument, self).__of__(self)

    def getAttributeNS(self, namespaceURI, localName):
        for item in self._attributes:
            if (  namespaceURI == item[_ATTR_NS]
                  and localName == item[_ATTR_LOCALNAME]):
                if type(item[_ATTR_VALUE]) in _StringTypes:
                    # value is a string
                    return item[_ATTR_VALUE]
                # value is a subtree
                return _attr_get_value(item[_ATTR_VALUE])
        for item in self._attr_info:
            if (  namespaceURI == item[_ATTR_NS]
                  and localName == item[_ATTR_LOCALNAME]):
                # will always be a string in the current implementation
                return item[_ATTR_VALUE]
        return ""

    def getAttributeNodeNS(self, namespaceURI, localName):
        for item in self._attributes:
            if (  namespaceURI == item[_ATTR_NS]
                  and localName == item[_ATTR_LOCALNAME]):
                return Attr(item, self.ownerDocument, self).__of__(self)
        for item in self._attr_info:
            if (  namespaceURI == item[_ATTR_NS]
                  and localName == item[_ATTR_LOCALNAME]):
                item = item[:]
                if self._attributes:
                    self._attributes.append(item)
                else:
                    self.__dict__['_attributes'] = [item]
                return Attr(item, self.ownerDocument, self).__of__(self)

    def hasAttribute(self, name):
        for item in self._attributes:
            if name == item[_ATTR_NAME]:
                return 1
        for item in self._attr_info:
            if name == item[_ATTR_NAME]:
                return 1
        return 0

    def hasAttributeNS(self, namespaceURI, localName):
        for item in self._attributes:
            if (  namespaceURI == item[_ATTR_NS]
                  and localName == item[_ATTR_LOCALNAME]):
                return 1
        for item in self._attr_info:
            if (  namespaceURI == item[_ATTR_NS]
                  and localName == item[_ATTR_LOCALNAME]):
                return 1
        return 0

    def removeAttribute(self, name):
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        for i in range(len(self._attributes)):
            item = self._attributes[i]
            if item[_ATTR_NAME] == name:
                break
        else:
            return
        if not self._attributes:
            self.__dict__['_attributes'] = []
        del self._attributes[i]
        if item[_ATTR_SPECIFIED]:
            self._changed()

    def removeAttributeNS(self, namespaceURI, localName):
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        for i in range(len(self._attributes)):
            item = self._attributes[i]
            if (item[_ATTR_NS] == namespaceURI
                and item[_ATTR_LOCALNAME] == localName):
                if not self._attributes:
                    self.__dict__['_attributes'] = []
                del self._attributes[i]
                self._changed()
                return

    def removeAttributeNode(self, oldAttr):
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        item = oldAttr._item
        if item in self._attributes:
            self._attributes.remove(item)
            oldAttr._set_owner_element(None)
            if item[_ATTR_SPECIFIED]:
                self._changed()
            return oldAttr
        else:
            raise xml.dom.NotFoundErr()

    # Because we don't _changed on any existing attr nodes, all attr nodes
    # must have the same persistent parent as their ownerElement
    def setAttribute(self, name, value):
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        if not _ok_qualified_name(name):
            raise xml.dom.InvalidCharacterErr()
        if type(value) not in _StringTypes:
            raise TypeError, "attribute value must be a string"
        if self._attributes:
            for item in self._attributes:
                if (  name == item[_ATTR_NAME]
                      and not item[_ATTR_NS]):
                    if type(item[_ATTR_VALUE]) in _StringTypes:
                        # value is a string
                        if item[_ATTR_VALUE] != value:
                            item[_ATTR_VALUE] = value
                            self._changed()
                    else:
                        # value is a list of children, there's an Attr node
                        # for this attr which shares this list.
                        if _attr_get_value(item[_ATTR_VALUE]) != value:
                            _attr_set_value(item, value)
                            self._changed()
                    return
        # attr hasn't been found
        if not self._attributes:
            self.__dict__['_attributes'] = []
        # should look up namespaceURI here...
        self._attributes.append([None, name, None, None, value, 1])
        self._changed()

    # Because we don't _changed on any existing attr nodes, all attr nodes
    # must have the same persistent parent as their ownerElement
    def setAttributeNS(self, namespaceURI, qualifiedName, value):
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        _check_qualified_name(qualifiedName, namespaceURI)
        if ":" in qualifiedName:
            prefix, localName = _string.split(qualifiedName, ":", 1)
            _check_reserved_prefixes(prefix, namespaceURI)
        elif namespaceURI:
            prefix = None
            localName = qualifiedName
        else:
            prefix = None
            localName = qualifiedName
        if self._attributes:
            # replace existing attribute rather than add new one
            for item in self._attributes:
                if (  namespaceURI == item[_ATTR_NS]
                      and localName == item[_ATTR_LOCALNAME]):
                    if type(item[_ATTR_VALUE]) in _StringTypes:
                        # value is a string
                        if (  item[_ATTR_VALUE] != value
                              or item[_ATTR_PREFIX] != prefix):
                            item[_ATTR_VALUE] = value
                            item[_ATTR_PREFIX] = prefix
                            item[_ATTR_NAME] = "%s:%s" % (
                                prefix, item[_ATTR_LOCALNAME])
                            self._changed()
                    else:
                        # value is a list of children, there's an Attr node
                        # for this attr which shares this list.
                        if (  _attr_get_value(item[_ATTR_VALUE]) != value
                              or item[_ATTR_PREFIX] != prefix):
                            _attr_set_value(item, value)
                            item[_ATTR_PREFIX] = prefix
                            item[_ATTR_NAME] = "%s:%s" % (
                                prefix, item[_ATTR_LOCALNAME])
                            self._changed()
                    return
        if not self._attributes:
            self.__dict__['_attributes'] = []
        self._attributes.append(
            [namespaceURI, qualifiedName, localName, prefix, value, 1])
        self._changed()

    def setAttributeNode(self, newAttr):
        return self._set_attribute_node(
            newAttr.name, _attr_item_match_name, newAttr)

    def setAttributeNodeNS(self, newAttr):
        name = (newAttr.namespaceURI, newAttr.localName)
        return self._set_attribute_node(
            name, _attr_item_match_ns, newAttr)

    def _set_attribute_node(self, name, matcher, newAttr):
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        if newAttr.ownerElement:
            raise xml.dom.InuseAttributeErr()
        if newAttr.nodeType != Node.ATTRIBUTE_NODE:
            raise xml.dom.HierarchyRequestErr(
                "attributes must have nodeType xml.dom.Node.ATTRIBUTE_NODE")
        if not self.ownerDocument.isSameNode(newAttr.ownerDocument):
            raise xml.dom.WrongDocumentErr()
        oldAttr = None
        if not self._attributes:
            self.__dict__['_attributes'] = []
        for i in range(len(self._attributes)):
            item = self._attributes[i]
            if matcher(item, name):
                # replace existing node # XXX set ownerElement for other nodes
                oldAttr = Attr(item, self.ownerDocument).__of__(self)
                self._attributes[i] = newAttr._item
                break
        if oldAttr is None:
            self._attributes.append(newAttr._item)
        newAttr._set_owner_element(self)
        self._changed()
        return oldAttr


class ChildNodeList:
    """NodeList implementation that provides the children of a single node.

    This is returned by Node.childNodes and Node._get_childNodes().
    The list operations supported by this object can be used to mutate
    the document contents.

    """
    def __init__(self, parent):
        self.__dict__['_parent'] = parent

    def __getstate__(self):
        raise RuntimeError, "ChildNodeList instances cannot be stored"

    def item(self, i):
        try:
            return self[i]
        except IndexError:
            return

    def __getitem__(self, i):
        return self._parent._children[i].__of__(self._parent)

    def __setitem__(self, i, newChild):
        p = self._parent
        oldChild = p._children[i].__of__(p)
        p.replaceChild(newChild, oldChild)

    def __delitem__(self, i):
        oldChild = self._parent._children[i].__of__(self._parent)
        self._parent.removeChild(oldChild)

    def _get_length(self):
        return len(self._parent._children)
    __len__ = _get_length

    def __getattr__(self, name):
        if name == "length":
            return len(self._parent._children)
        raise AttributeError, name

    def __setattr__(self, name, value):
        if name == "length":
            raise xml.dom.NoModificationAllowedErr()
        else:
            raise TypeError, "NodeList has only read-only attributes"

    def __nonzero__(self):
        return self._parent._children and 1 or 0

    def count(self, value):
        value = _aq_base(value)
        for node in self._parent._children:
            if node is value:
                return 1
        return 0

    def index(self, value):
        children = self._parent._children
        value = _aq_base(value)
        for i in range(len(children)):
            if value is children[i]:
                return i
        raise ValueError, "NodeList.index(x): x not in sequence"


class SimpleNodeList:
    """NodeList implementation that contains pre-wrapped nodes.

    This is returned by the getElementsByTagName() and
    getElementsByTagNameNS() methods of Document and Element.
    It cannot be used to mutate the document contents.

    """
    def __init__(self, list=None):
        if list is None:
            list = []
        self.__dict__['_data'] = list

    def __getstate__(self):
        raise RuntimeError, "SimpleNodeList instances cannot be stored"

    #
    #  NodeList interface
    #

    def __getattr__(self, name):
        if name == "length":
            return len(self._data)
        raise AttributeError, name

    def __setattr__(self, name, value):
        if name == "length":
            raise xml.dom.NoModificationAllowedErr()
        self.__dict__[name] = value

    def _get_length(self):
        return len(self._data)

    def item(self, i):
        if 0 <= i < len(self._data):
            return self._data[i]

    #
    #  Read-only sequence interface
    #

    def __contains__(self, obj):
        base = _aq_base(obj)
        for node in self._data:
            if _aq_base(node) is base:
                return 1
        return 0

    def __getitem__(self, i):
        return self._data[i]

    def __getslice__(self, i, j):
        return SimpleNodeList(self._data[i:j])

    def __len__(self):
        return len(self._data)

    def __nonzero__(self):
        return self._data and 1 or 0

    def count(self, value):
        value = _aq_base(value)
        for node in self._data:
            if _aq_base(node) is value:
                return 1
        return 0

    def index(self, value):
        value = _aq_base(value)
        for i in range(len(self._data)):
            base = _aq_base(self._data[i])
            if value is base:
                return i
        raise ValueError, "NodeList.index(x): x not in sequence"


class CharacterData(Childless, Node):
    def __init__(self, data):
        d = self.__dict__
        d['_in_tree'] = 0
        d['data'] = data
        d['nodeValue'] = data

    def _get_data(self):
        return self.data

    def _set_data(self, data):
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        d = self.__dict__
        if d['data'] != data:
            d['data'] = data
            d['nodeValue'] = data
            self._changed()

    _set_nodeValue = _set_data

    def _get_length(self):
        return len(self.data)
    length = ComputedAttribute(_get_length, 1)

    def __len__(self):
        return len(self.data)

    def appendData(self, arg):
        if arg:
            data = self.data + arg
            d = self.__dict__
            d['data'] = data
            d['nodeValue'] = data
            self._changed()

    def deleteData(self, offset, count):
        if count < 0 or offset < 0 or offset > len(self.data):
            raise xml.dom.IndexSizeErr()
        if count:
            data = self.data[:offset] + self.data[offset+count:]
            self.__dict__['data'] = data
            self.__dict__['nodeValue'] = data
            self._changed()

    def insertData(self, offset, arg):
        if offset < 0 or offset > len(self.data):
            raise xml.dom.IndexSizeErr()
        if arg:
            data = self.data
            data = _string.join((data[:offset], arg, data[offset:]), '')
            self.__dict__['data'] = data
            self.__dict__['nodeValue'] = data
            self._changed()

    def replaceData(self, offset, count, arg):
        if count < 0 or offset < 0 or offset > len(self.data):
            raise xml.dom.IndexSizeErr()
        if count or arg:
            data = self.data
            data = _string.join((data[:offset], arg, data[offset+count:]), '')
            self.__dict__['data'] = data
            self.__dict__['nodeValue'] = data
            self._changed()

    def substringData(self, offset, count):
        if count < 0 or offset < 0 or offset > len(self.data):
            raise xml.dom.IndexSizeErr()
        return self.data[offset:offset+count]

    # DOM Level 3 (working draft, 5 June 2001)

    def _get_textContent(self):
        return self.nodeValue
    textContent = ComputedAttribute(_get_textContent, 1)


class Text(CharacterData):
    nodeName = "#text"
    nodeType = Node.TEXT_NODE

    def splitText(self, offset):
        if offset < 0 or offset > len(self.data):
            raise xml.dom.IndexSizeErr()
        parent = self.parentNode
        newText = Text(self.data[offset:])
        data = self.data[:offset]
        self.__dict__['data'] = self.__dict__['nodeValue'] = data
        if parent is not None:
            newText = newText.__of__(parent)
            sibmap = parent._v_sibling_map
            if sibmap:
                del self.__dict__['_v_sibling_map']
                #prev, next = sibmap[_aq_base(self)]
                #if next is not None:
                #    next = next.__of__(parent)
            else:
                next = self.nextSibling
            if next is None:
                parent.appendChild(newText)
            else:
                parent.insertBefore(newText, next.__of__(parent))
        self._changed()
        return newText

    # DOM Level 3 (working draft 01 Sep 2000)
    isWhitespaceInElementContent = 0
    def _get_isWhitespaceInElementContent(self):
        return self.isWhitespaceInElementContent


class Comment(CharacterData):
    nodeName = "#comment"
    nodeType = Node.COMMENT_NODE


# Attr nodes can share their _children with the attribute structure of
# their ownerElements, so this list reference must never be changed - don't
# replace _children, instead add and remove list members.  Same for _item
# members, which are shared with the ownerElement and other Attr ndoes.
#
# We expect that the usual access of attrs is via the element string methods
# getAttribute* and setAttribute, so usually attr values are stored as strings
# in the element.  When getAttributeNode is called, we turn the string into a
# list with a single text node, the attr node shares this reference.  Similarly
# for setAttributeNode.  We could stay with the string in many cases at the
# cost of complexity.
class Attr(Parentless, Node):
    nodeType = Node.ATTRIBUTE_NODE

    _in_tree = 0

    _allowed_child_types = (Node.TEXT_NODE,
                            Node.ENTITY_REFERENCE_NODE)

    def __init__(self, item, ownerDocument, ownerElement=None):
        d = self.__dict__
        # attributes that must be shared with the ownerElement's representation
        if type(item[_ATTR_VALUE]) in _StringTypes:
            # turn string representation into list of children
            itemNode = Text(item[_ATTR_VALUE])
            del itemNode.__dict__['_in_tree']
            item[_ATTR_VALUE] = [itemNode]
        d['_children'] = item[_ATTR_VALUE]
        d['_item'] = item
        # attributes that aren't shared with the ownerElement
        d['ownerDocument'] = ownerDocument
        if item[_ATTR_NS]:
            d['namespaceURI'] = item[_ATTR_NS]
            d['localName'] = item[_ATTR_LOCALNAME]
        # ownerElement arg is for readonlyness, not the OwnerElement attribute
        if ownerElement is not None and ownerElement._readonly:
            d['_readonly'] = 1 # XXX must be shared in case of removal?
        d['specified'] = item[_ATTR_SPECIFIED]

    def __getstate__(self):
        raise RuntimeError("Attr nodes cannot be pickled")

    def __cmp__(self, other):
        if (other.nodeType == Node.ATTRIBUTE_NODE
            and self._item is other._item):
            return 0
        else:
            return cmp(id(_aq_base(self)), id(_aq_base(other)))

    def __repr__(self):
        return "<Attr '%s' at 0x%x; identity=0x%x>" % (
            self.name, id(_aq_base(self)), id(self._item))

    def _cloneNode(self, deep, mutable, document):
        # self is *not* an acquisition wrapper!
        clone = self.__class__.__basicnew__()
        d = clone.__dict__
        d.update(self.__dict__)
        d['_item'] = item = list(self._item)
        item[_ATTR_SPECIFIED] = 1
        d['specified'] = 1
        if d.has_key('_readonly'):
            del d['_readonly']
        # clone the children
        d['_children'] = L = []
        for child in self._children:
            newChild = child._cloneNode(1, mutable, document)
            if newChild.__dict__.has_key('_in_tree'):
                del newChild.__dict__['_in_tree']
            L.append(newChild)
        return clone

    #nodeName, name, prefix must be shared between attr nodes and
    #element storage, because we can change the prefix with either interface.

    _get_name = Node._get_nodeName

    def _get_nodeName(self):
        return self._item[_ATTR_NAME]        
    nodeName = ComputedAttribute(_get_nodeName, 1)

    def _get_name(self):
        return self._item[_ATTR_NAME]        
    name = ComputedAttribute(_get_name, 1)

    def _get_prefix(self):
        return self._item[_ATTR_PREFIX]
    prefix = ComputedAttribute(_get_prefix, 1)

    # we aren't checking to see if this attr was created with a lvl2 method;
    # if it wasn't, setting the prefix will make the name funny -
    # but that's undefined behavior anyway.
    def _set_prefix(self, value):
        self._check_prefix(value)
        d = self.__dict__
        self._item[_ATTR_PREFIX] = value
        if value:
            name =  "%s:%s" % (value, self._item[_ATTR_LOCALNAME])
        else:
            name = self._item[_ATTR_LOCALNAME]
        self._item[_ATTR_NAME] = name
        if not self.specified:
            self.__dict__['specified'] = 1
            self._item[_ATTR_SPECIFIED] = 1
            self._changed()

    def _set_owner_element(self, owner):
        # set ownerElement with acquisition XXX same aq bug as parent usage
        _reparent(self, owner) 
        if owner is not None:
            self.__dict__['ownerDocument'] = owner.ownerDocument

    def _set_owner_document(self, doc):
        self.__dict__['ownerDocument'] = doc

    def _get_ownerDocument(self):
        return self.ownerDocument

    def _get_ownerElement(self):
        # Acquire the owner.
        parent = _parent_of(self)
        if parent and parent.isSameNode(self.ownerDocument):
            return None                 # we use this for unowned attrs
        return parent
    ownerElement = ComputedAttribute(_get_ownerElement, 1)

    def _get_value(self):
        return _attr_get_value(self._children)
    value = ComputedAttribute(_get_value)

    _get_nodeValue = _get_value
    nodeValue = value

    def _set_value(self, data):
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        _attr_set_value(self._item, data)        

    _set_nodeValue = _set_value

    childNodes = ComputedAttribute(Node._get_childNodes)

    # DOM Level 3 (Working Draft, 01 Sep 2000)

    def _get_specified(self):
        return self.specified

    def _get_textContent(self):
        return self._item[ATTR_VALUE]
    textContent = ComputedAttribute(_get_textContent)

    def isSameNode(self, other):
        return (other is not None
                and other.nodeType == Node.ATTRIBUTE_NODE
                and self._item is other._item)


class MapFromParent(AttributeControl):
    """
    Baseclass for a NamedNodeMap that works by extracting information
    from a parent.
    Must be subclassed to determine what we're looking for and returning.
    """

    def __init__(self, parent):
        self.__dict__['_parent'] = parent
        # subclass must set _parentListName

    def _getParentList(self):
        "return the parent's list used to make this map"
        return getattr(self._parent, self._parentListName)

    def __getstate__(self):
        raise RuntimeError, "NamedNodeMap type instances cannot be stored"

    def _get_length(self):
        return len(self._getParentList())
    __len__ = _get_length

    def __getattr__(self, name):
        if name == "length":
            return self._get_length()
        raise AttributeError, name

    def __setattr__(self, name, value):
        if name == "length":
            raise xml.dom.NoModificationAllowedErr()
        AttributeControl.__setattr__(self, name, value)

    def get(self, name, default=None):
        node = self.getNamedItem(name)
        if node is None:
            return default
        else:
            return node

    def item(self, i):
        try:
            itemSource = self._getParentList()[i]
        except IndexError:
            return
        else:
            return self._item_helper(itemSource)

    # subclass must define _item_helper

    def getNamedItem(self, name):
        for item in self._getParentList():
            if self._nameMatcher(item, name):
                return self._item_helper(item)

    def getNamedItemNS(self, namespaceURI, localName):
        for item in self._getParentList():
            if self._nsMatcher(item, (namespaceURI, localName)):
                return self._item_helper(item)

    def __getitem__(self, name):
        node = self.getNamedItem(name)
        if node is None:
            raise KeyError, name
        return node

    # subclass must define _set_named_item, _nameMatcher, nsMatcher

    def setNamedItem(self, node):
        return self._set_named_item(node.nodeName, self._nameMatcher, node)

    def setNamedItemNS(self, node):
        nameinfo = (node.namespaceURI, node.localName)
        return self._set_named_item(nameinfo, self._nsMatcher, node)

    def __setitem__(self, name, node):
        if self._parent._readonly or self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        assert name == node.nodeName
        self.setNamedItem(node)

    # subclass must define _key_helper, _delFromParentList

    def removeNamedItem(self, name):
        return self._remove_named_item(name, self._nameMatcher)

    def removeNamedItemNS(self, namespaceURI, localName):
        return self._remove_named_item((namespaceURI, localName),
                                       self._nsMatcher)

    def _remove_named_item(self, name, matcher):
        #
        # The workhorse of item removal; this removes whatever
        # item the 'matcher' test determines matches.  'name' is
        # passed to the matcher but is not used otherwise.
        #
        if self._parent._readonly or self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        pList = self._getParentList()
        for i in range(len(pList)):
            item = pList[i]
            if matcher(item, name):
                break
        else:
            raise xml.dom.NotFoundErr()
        self._delFromParentList(pList, i)        
        node = self._item_helper(item)
        node._set_owner_element(None)
        return node

    def __delitem__(self, name):
        if self._parent._readonly or self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        pList = self._getParentList()
        for i in range(len(pList)):
            item = pList[i]
            s = self._key_helper(item)
            if s == name:
                self._delFromParentList(pList, i)
                return
        raise KeyError, name

    def has_key(self, name):
        for item in self._getParentList():
            if self._key_helper(item) == name:
                return 1
        return 0

    def items(self):
        L = []
        for item in self._getParentList():
            L.append((self._key_helper(item), self._item_helper(item)))
        return L

    def keys(self):
        L = []
        for item in self._getParentList():
            L.append(self._key_helper(item))
        return L

    def values(self):
        L = []
        for item in self._getParentList():
            L.append(self._item_helper(item))
        return L



class AttributeMap(MapFromParent):
    """NamedNodeMap that works on the attribute structure.

    This doesn't do anything about the namespace declarations.
    """

    _parentListName = '_attributes'

    def __init__(self, parent):
        d = self.__dict__
        d['_attr_info'] = parent._attr_info
        d['_parent'] = parent
        d['_nameMatcher'] = _attr_item_match_name
        d['_nsMatcher'] = _attr_item_match_ns

    def _get_length(self):
        d = {}
        for item in self._parent._attributes:
            if item[_ATTR_NS]:
                key = item[_ATTR_NS], item[_ATTR_LOCALNAME]
            else:
                key = item[_ATTR_NAME]
            d[key] = 1
        for item in self._attr_info:
            if item[_ATTR_NS]:
                key = item[_ATTR_NS], item[_ATTR_LOCALNAME]
            else:
                key = item[_ATTR_NAME]
            d[key] = 1
        return len(d)
    __len__ = _get_length

    def item(self, i):
        node = MapFromParent.item(self, i)
        if node is None and self._attr_info:
            d = {}
            for item in self._parent._attributes:
                if item[_ATTR_NS]:
                    key = item[_ATTR_NS], item[_ATTR_LOCALNAME]
                else:
                    key = item[_ATTR_NAME]
                d[key] = 1
            j = len(d)
            for item in self._attr_info:
                name = item[_ATTR_NAME]
                if d.has_key(name):
                    pass
                else:
                    if j == i:
                        item = list(item)
                        if self._parent._attributes:
                            self._parent._attributes.append(item)
                        else:
                            self._parent.__dict__['_attributes'] = [item]
                        node = Attr(
                            item, self._parent.ownerDocument, self._parent)
                        node = node.__of__(self._parent)
                        break
                    j = j + 1
        return node

    def _item_helper(self, itemSource):
        "used by item; create an Attribute from the item and return it"
        node = Attr(itemSource, self._parent.ownerDocument, self._parent)
        return node.__of__(self._parent)

    def _set_named_item(self, nameinfo, matcher, node):
        "utility function for setNamedItem"
        if self._parent._readonly or self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        if node.nodeType != Node.ATTRIBUTE_NODE:
            raise xml.dom.HierarchyRequestErr()
        if not self._parent.ownerDocument.isSameNode(node.ownerDocument):
            raise xml.dom.WrongDocumentErr()
        if node.ownerElement:
            if node.ownerElement.isSameNode(self._parent):
                # This is already our node; no extra work needed, and no
                # change to the storage object.
                return node
            raise xml.dom.InuseAttributeErr()
        attributes = self._getParentList()
        if not attributes:
            self._parent.__dict__['_attributes'] = [node._item]
            node._set_owner_element(self._parent)
            return node
        oldNode = None
        for i in range(len(attributes)):
            item = attributes[i]
            if matcher(item, nameinfo):
                oldNode = item
                attributes[i] = node._item
                break
        if oldNode is None:
            self._addToParentList(attributes, node)
            node._set_owner_element(self._parent) 
        return oldNode

    def _delFromParentList(self, attrs, i):
        "workhorse for __delitem__; remove ith item from attrs"
        del attrs[i] #XXX ownerElement needs to be updated in other refs
        self._parent._changed()

    def _addToParentList(self, attrs, node):
        if self._parent._attributes:
            self._parent._attributes.append(node._item)
        else:
            self._parent.__dict__['_attributes'] = [node._item]
        self._parent._changed()

    def _key_helper(self, itemSource):
        "given an item source, return an appropriate key for our mapping"
        return itemSource[_ATTR_NAME]


# Utility functions for Attrs, used by more than the Attr class.

def _attr_item_match_name(item, name,
                          _ATTR_NAME=_ATTR_NAME):
    "utility function for AttributeMap; return true if name matches item"
    return item[_ATTR_NAME] == name

def _attr_item_match_ns(item, (namespaceURI, localName),
                        _ATTR_NS=_ATTR_NS, _ATTR_LOCALNAME=_ATTR_LOCALNAME):
    "utility function for AttributeMap; return true if name matches item"    
    return (item[_ATTR_LOCALNAME] == localName
            and item[_ATTR_NS] == namespaceURI)

def _attr_get_value(nodes):
    "utility function to get attr value; concatenate values of list of nodes"
    L = []
    for node in nodes:
        L.append(node.nodeValue)
    return _string.join(filter(None, L), '')

def _attr_set_value(item, value):
    "utility function to safely set shared value of attr item"
    newChild = Text(value)
    del newChild.__dict__['_in_tree']
    while item[_ATTR_VALUE]:
        item[_ATTR_VALUE].pop()
    item[_ATTR_VALUE].append(newChild)


# no longer needed
del ComputedAttribute



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

import os, sys
  
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing import ZopeTestCase

ZopeTestCase.installProduct('ParsedXML')

from Products.ParsedXML import DOM
from Products.ParsedXML.DOM import ExpatBuilder
from StringIO import StringIO
from domapi import DOMImplementationTestSuite

def DOMParseString(xml):
    file = StringIO(xml)
    return ExpatBuilder.parse(file)

def get_element_ids(node):
    result = []
    if node.nodeType == node.ELEMENT_NODE:
        result.append(node.elementId)
    for child in node.childNodes:
        result.extend(get_element_ids(child))
    return result

def check_unique_ids(node):
    element_ids = get_element_ids(node)
    element_ids.sort()
    last = element_ids[0]
    for element_id in element_ids[1:]:
        assert last != element_id,\
               'Found at least one double element_id: %s' % element_id
        last = element_id

class ElementIdTestCase(ZopeTestCase.ZopeTestCase):
    def afterSetUp(self):
        self.doc = DOMParseString('''
        <doc>
        <p>Test</p>
        <p>Another test</p>
        <p>Foo<sub/><sub2>hey</sub2></p>
        </doc>
        ''')

    def testElementIdsAfterParse(self):
        element_ids = get_element_ids(self.doc)
        for i in xrange(len(element_ids)):
            assert i == element_ids[i]
            
    def testElementIdsAfterAppend(self):
        element = self.doc.createElement('foo')
        self.doc.documentElement.appendChild(element)
        check_unique_ids(self.doc)

    def testElementIdsAfterAppend2(self):
        element = self.doc.createElement('foo')
        self.doc.documentElement.insertBefore(
            element, self.doc.documentElement.childNodes[0])
        check_unique_ids(self.doc)

    def testCloneNodeShallow(self):
        cloned = self.doc.documentElement.cloneNode(0)
        self.doc.documentElement.appendChild(cloned)
        check_unique_ids(self.doc)

    def testCloneNodeDeep(self):
        cloned = self.doc.documentElement.cloneNode(1)
        self.doc.documentElement.appendChild(cloned)
        check_unique_ids(self.doc)

    def testImportNodeShallow(self):
        otherdoc = DOMParseString('''
        <hey><some></some></hey>
        ''')
        imported = self.doc.importNode(
            otherdoc.documentElement, 0)
        self.doc.documentElement.appendChild(imported)
        check_unique_ids(self.doc)
        
    def testImportNodeDeep(self):
        otherdoc = DOMParseString('''
        <hey><some></some></hey>
        ''')
        imported = self.doc.importNode(
            otherdoc.documentElement, 1)
        self.doc.documentElement.appendChild(imported)
        check_unique_ids(self.doc)
        
if __name__ == '__main__':
    framework()
else:
    import unittest
    def test_suite():
        suite = unittest.TestSuite()
        suite.addTest(unittest.makeSuite(ElementIdTestCase))
        return suite



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Implementation of the DOM Level 3 'Load' feature."""

import Core
import ExpatBuilder

import copy
import string
import xml.dom


__all__ = ["DOMBuilder", "DOMEntityResolver", "DOMInputSource"]


class DOMBuilder(Core.AttributeControl):
    entityResolver = None
    errorHandler = None
    filter = None

    def __init__(self):
        self.__dict__['_options'] = ExpatBuilder.Options()

    def _get_entityResolver(self):
        return self.entityResolver
    def _set_entityResolver(self, entityResolver):
        self.__dict__['entityResolver']

    def _get_errorHandler(self):
        return self.errorHandler
    def _set_errorHandler(self, errorHandler):
        self.__dict__['errorHandler'] = errorHandler

    def _get_filter(self):
        return self.filter
    def _set_filter(self, filter):
        self.__dict__['filter'] = filter

    def setFeature(self, name, state):
        if self.supportsFeature(name):
            try:
                settings = self._settings[(_name_xform(name), state)]
            except KeyError:
                raise xml.dom.NotSupportedErr(
                    "unsupported feature: " + `name`)
            else:
                for name, value in settings:
                    setattr(self._options, name, value)
        else:
            raise xml.dom.NotFoundErr("unknown feature: " + `name`)        

    def supportsFeature(self, name):
        return hasattr(self._options, _name_xform(name))

    def canSetFeature(self, name, state):
        key = (_name_xform(name), state and 1 or 0)
        return self._settings.has_key(key)

    _settings = {
        ("namespaces", 0): [("namespaces", 0)],
        ("namespaces", 1): [("namespaces", 1)],
        ("namespace_declarations", 0): [("namespace_declarations", 0)],
        ("namespace_declarations", 1): [("namespace_declarations", 1)],
        ("validation", 0): [("validation", 0)],
        ("external_general_entities", 0): [("external_general_entities", 0)],
        ("external_general_entities", 1): [("external_general_entities", 1)],
        ("external_parameter_entities", 0): [("external_parameter_entities", 0)],
        ("external_parameter_entities", 1): [("external_parameter_entities", 1)],
        ("validate_if_cm", 0): [("validate_if_cm", 0)],
        ("create_entity_ref_nodes", 0): [("create_entity_ref_nodes", 0)],
        ("create_entity_ref_nodes", 1): [("create_entity_ref_nodes", 1)],
        ("entity_nodes", 0): [("create_entity_ref_nodes", 0),
                              ("entity_nodes", 0)],
        ("entity_nodes", 1): [("entity_nodes", 1)],
        ("white_space_in_element_content", 0):
            [("white_space_in_element_content", 0)],
        ("white_space_in_element_content", 1):
            [("white_space_in_element_content", 1)],
        ("cdata_nodes", 0): [("cdata_nodes", 0)],
        ("cdata_nodes", 1): [("cdata_nodes", 1)],
        ("comments", 0): [("comments", 0)],
        ("comments", 1): [("comments", 1)],
        ("charset_overrides_xml_encoding", 0):
            [("charset_overrides_xml_encoding", 0)],
        ("charset_overrides_xml_encoding", 1):
            [("charset_overrides_xml_encoding", 1)],
    }

    def getFeature(self, name):
        try:
            return getattr(self._options, _name_xform(name))
        except AttributeError:
            raise xml.dom.NotFoundErr()

    def parseURI(self, uri):
        if self.entityResolver:
            input = self.entityResolver.resolveEntity(None, uri)
        else:
            input = DOMEntityResolver().resolveEntity(None, uri)
        return self.parseDOMInputSource(input)

    def parseDOMInputSource(self, input):
        options = copy.copy(self._options)
        options.filter = self.filter
        options.errorHandler = self.errorHandler
        fp = input.byteStream
        if fp is None and options.systemId:
            import urllib
            fp = urllib.urlopen(input.systemId)
        builder = ExpatBuilder.makeBuilder(options)
        return builder.parseFile(fp)


class DOMEntityResolver(Core.DOMImplementation):
    def resolveEntity(self, publicId, systemId):
        source = DOMInputSource()
        source.publicId = publicId
        source.systemId = systemId
        if systemId:
            import urllib
            self.byteStream = urllib.urlopen(systemId)
            # Should parse out the content-type: header to
            # get charset information so that we can set the
            # encoding attribute on the DOMInputSource.
        return source


class DOMInputSource(Core.AttributeControl):
    byteStream = None
    characterStream = None
    encoding = None
    publicId = None
    systemId = None

    def _get_byteStream(self):
        return self.byteStream
    def _set_byteStream(self, byteStream):
        self.__dict__['byteStream'] = byteStream

    def _get_characterStream(self):
        return self.characterStream
    def _set_characterStream(self, characterStream):
        self.__dict__['characterStream'] = characterStream

    def _get_encoding(self):
        return self.encoding
    def _set_encoding(self, encoding):
        self.__dict__['encoding'] = encoding

    def _get_publicId(self):
        return self.publicId
    def _set_publicId(self, publicId):
        self.__dict__['publicId'] = publicId

    def _get_systemId(self):
        return self.systemId
    def _set_systemId(self, systemId):
        self.__dict__['systemId'] = systemId


class DOMBuilderFilter:
    """Element filter which can be used to tailor construction of
    a DOM instance.
    """

    # There's really no need for this class; concrete implementations
    # should just implement the endElement() method as appropriate.

    def endElement(self, element):
        # Why this method is supposed to return anything at all
        # is a mystery; the result doesn't appear to be used.
        return 1


def _name_xform(name):
    return string.replace(string.lower(name), '-', '_')



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Constructor for ParsedXML.DOM, based on a SAX parser."""

import os
import urllib
import xml.sax


class SAXBuilder(xml.sax.ContentHandler):
    _locator = None
    document = None
    documentElement = None

    def __init__(self, documentFactory=None):
        self.documentFactory = documentFactory
        self._ns_contexts = [{}] # contains uri -> prefix dicts
        self._current_context = self._ns_contexts[-1]

    def install(self, parser):
        parser.setContentHandler(self)

    def setDocumentLocator(self, locator):
        self._locator = locator

    def startPrefixMapping(self, prefix, uri):
        self._ns_contexts.append(self._current_context.copy())
        self._current_context[uri] = prefix or None

    def endPrefixMapping(self, prefix):
        self._current_context = self._ns_contexts.pop()

    def _make_qname(self, uri, localname, tagname):
        # When using namespaces, the reader may or may not
        # provide us with the original name. If not, create
        # *a* valid tagName from the current context.
        if uri:
            if tagname is None:
                prefix = self._current_context.get(uri)
                if prefix:
                    tagname = "%s:%s" % (prefix, localname)
                else:
                    tagname = localname
        else:
            tagname = localname
        return tagname

    def startElementNS(self, name, tagName, attrs):
        uri, localname = name
        tagName = self._make_qname(uri, localname, tagName)
        if not self.document:
            factory = self.documentFactory
            self.document = factory.createDocument(uri or None, tagName, None)
            node = self.document.documentElement
        else:
            if uri:
                node = self.document.createElementNS(uri, tagName)
            else:
                node = self.document.createElement(localname)
            self.curNode.appendChild(node)
        self.curNode = node

        for aname, value in attrs.items():
            a_uri, a_localname = aname
            if a_uri:
                qname = "%s:%s" % (self._current_context[a_uri], a_localname)
                node.setAttributeNS(a_uri, qname, value)
            else:
                attr = self.document.createAttribute(a_localname)
                node.setAttribute(a_localname, value)

    def endElementNS(self, name, tagName):
        self.curNode = self.curNode.parentNode

    def startElement(self, name, attrs):
        if self.documentElement is None:
            factory = self.documentFactory
            self.document = factory.createDocument(None, name, None)
            node = self.document.documentElement
            self.documentElement = 1
        else:
            node = self.document.createElement(name)
            self.curNode.appendChild(node)
        self.curNode = node

        for aname, value in attrs.items():
            node.setAttribute(aname, value)

    def endElement(self, name):
        self.curNode = self.curNode.parentNode

    def comment(self, s):
        node = self.document.createComment(s)
        self.curNode.appendChild(node)

    def processingInstruction(self, target, data):
        node = self.document.createProcessingInstruction(target, data)
        self.curNode.appendChild(node)

    def ignorableWhitespace(self, chars):
        node = self.document.createTextNode(chars)
        self.curNode.appendChild(node)

    def characters(self, chars):
        node = self.document.createTextNode(chars)
        self.curNode.appendChild(node)


def parse(file, namespaces=1, dom=None, parser=None):
    if not parser:
        parser = xml.sax.make_parser()
    parser.setFeature(xml.sax.handler.feature_namespaces, namespaces)
    if not dom:
        import Core
        dom = Core.theDOMImplementation
    if isinstance(file, type('')):
        try:
            fp = open(file)
        except IOError, e:
            if e.errno != errno.ENOENT:
                raise
            fp = urllib.urlopen(file)
            systemId = file
        else:
            # Ugh!  Why doesn't urllib.pathname2url() do something useful?
            systemId = "file://" + os.path.abspath(file)
    else:
        source = xml.sax.InputSource()
        fp = file
        try:
            systemId = file.name
        except AttributeError:
            systemId = None
    source = xml.sax.InputSource(file)
    source.setByteStream(fp)
    source.setSystemId(systemId)
    builder = SAXBuilder(documentFactory=dom)
    builder.install(parser)
    parser.parse(source)
    if fp is not file:
        fp.close()
    return builder.document



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Test that some Zope interfaces are supported properly."""

import os, sys
  
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing import ZopeTestCase

ZopeTestCase.installProduct('ParsedXML')

from Products.ParsedXML import ParsedXML
from StringIO import StringIO

def assertSize(doc):
    "assert that the document size what's reported by len"
    gs = doc.get_size()
    l = len(str(doc))
    assert gs == l, "get_size reports %d while len reports %d" % (gs, l)
    
class GetSizeTestCase(ZopeTestCase.ZopeTestCase):
    "test that get_size works.  We only test on the persistent Document."

    def setUp(self):
        self.document = ParsedXML.ParsedXML('foo')

    def testGetSize(self):
        "assert that get_size works"
        assertSize(self.document)

    def testGetSizeParse(self):
        "assert that get_size works after a parse"        
        inStr = '<spam><eggs attr1="foo"><ham/>text</eggs></spam>'
        self.document.parseXML(StringIO(inStr))
        assertSize(self.document)        
        self.document.documentElement.parseXML(StringIO(inStr))
        assertSize(self.document)        

    def testGetSizeDOMMethods(self):
        "assert that get_size works after some DOM method manipulations"
        self.document.documentElement.appendChild(
            self.document.createElement('spam'))
        assertSize(self.document)
        self.document.documentElement.appendChild(
            self.document.createTextNode('spam'))
        assertSize(self.document)
        self.document.documentElement.appendChild(
            self.document.createTextNode('spam'))
        assertSize(self.document)
        self.document.normalize()
        assertSize(self.document)
        self.document.documentElement.setAttribute('eggs', 'ham')
        assertSize(self.document)

    def testGetSizeDOMAttributes(self):
        "assert that get_size works after some DOM attribute manipulations"
        self.document.documentElement.appendChild(
            self.document.createTextNode('spam'))
        self.document.documentElement.firstChild.data = "spamspamspam"
        assertSize(self.document)
        self.document.documentElement.setAttribute('eggs', 'ham')
        self.document.documentElement.attributes.item(0).value = 'spam'
        assertSize(self.document)

if __name__ == '__main__':
    framework()
else:
    import unittest
    def test_suite():
        suite = unittest.TestSuite()
        suite.addTest(unittest.makeSuite(GetSizeTestCase))
        return suite
    



##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################

import xml.dom
from xml.dom import Node

import sys
import unittest

# Namespace URI for namespace tests
TEST_NAMESPACE = 'uri:namespacetests'

# Convenience map for assert messages
TYPE_NAME = {
        Node.ATTRIBUTE_NODE:                'Attribute',
        Node.CDATA_SECTION_NODE:            'CDATA Section',
        Node.COMMENT_NODE:                  'Comment',
        Node.DOCUMENT_FRAGMENT_NODE:        'Document Fragment',
        Node.DOCUMENT_NODE:                 'Document',
        Node.DOCUMENT_TYPE_NODE:            'DocumentType',
        Node.ELEMENT_NODE:                  'Element',
        Node.ENTITY_NODE:                   'Entity',
        Node.ENTITY_REFERENCE_NODE:         'Entity Reference',
        Node.NOTATION_NODE:                 'Notation',
        Node.PROCESSING_INSTRUCTION_NODE:   'Processing Instruction',
        Node.TEXT_NODE:                     'Text',
   }

def checkAttribute(node, attribute, value):
    """Check that an attribute holds the expected value, and that the
    corresponding accessor method, if provided, returns an equivalent
    value."""
    #
    v1 = getattr(node, attribute)
    if v1 != value:
        raise AssertionError(
            "attribute value does not match\n  expected: %s\n  found: %s"
            % (`value`, `v1`))
    if hasattr(node, "_get_" + attribute):
        v2 = getattr(node, "_get_" + attribute)()
        if v2 != value:
            raise AssertionError(
                "accessor result does not match\n  expected: %s\n  found: %s"
                % (`value`, `v2`))
        if v1 != v2:
            raise AssertionError(
                "attribute & accessor result don't compare equal\n"
                "  attribute: %s\n  accessor: %s"
                % (`v1`, `v2`))


def checkAttributeNot(node, attribute, value):
    """Check that an attribute doesn't hold a specific failing value,
    and that the corresponding accessor method, if provided, returns
    an equivalent value."""
    #
    v1 = getattr(node, attribute)
    if v1 == value:
        raise AssertionError(
           "attribute value should not match\n  found: %s" % `v1`)
    if hasattr(node, "_get_" + attribute):
        v2 = getattr(node, "_get_" + attribute)()
        if v2 == value:
            raise AssertionError(
                "accessor result should not match\n  found: %s" % `v2`)
        if v1 != v2:
            raise AssertionError(
                "attribute & accessor result don't compare equal\n"
                "  attribute: %s\n  accessor: %s"
                % (`v1`, `v2`))

def checkAttributeSameNode(node, attribute, value):
    v1 = getattr(node, attribute)
    if value is None:
        if v1 is not None:
            raise AssertionError(
                "attribute value does not match\n  expected: %s\n  found: %s"
                % (`value`, `v1`))
    else:
        if not isSameNode(value, v1):
            raise AssertionError(
                "attribute value does not match\n  expected: %s\n  found: %s"
                % (`value`, `v1`))
    if hasattr(node, "_get_" + attribute):
        v2 = getattr(node, "_get_" + attribute)()
        if value is None:
            if v2 is not None:
                raise AssertionError(
                    "accessor result does not match\n"
                    "  expected: %s\n  found: %s" % (`value`, `v2`))
        elif not isSameNode(value, v2):
            raise AssertionError(
                "accessor result does not match\n"
                "  expected: %s\n  found: %s" % (`value`, `v2`))


def checkReadOnly(node, attribute):
    try:
        setattr(node, attribute, "don't set this!")
    except xml.dom.NoModificationAllowedErr:
        pass
    else:
        raise AssertionError("write-access to the '%s' attribute not blocked"
                             % attribute)

    if hasattr(node, "_set_" + attribute):
        # setter implemented; make sure it won't allow update
        try:
            getattr(node, "_set_" + attribute)("don't set this!")
        except xml.dom.NoModificationAllowedErr:
            pass
        else:
            raise AssertionError("_set_%s() allowed attribute update"
                                 % attribute)


def checkLength(node, value):
    checkAttribute(node, "length", value)
    if len(node) != value:
        raise AssertionError("broken support for __len__()")
    checkReadOnly(node, "length")


def isSameNode(node1, node2):
    """Compare two nodes, returning true if they are the same.

    Use the DOM lvl 3 Node.isSameNode method if available, otherwise use a
    simple 'is' test.

    """

    if hasattr(node1, 'isSameNode'):
        return node1.isSameNode(node2)
    else:
        return node1 is node2


class TestCaseBase(unittest.TestCase):

    def createDocumentNS(self):
        self.document = self.implementation.createDocument(
            TEST_NAMESPACE, 'foo:bar', None)
        return self.document

    def createDocument(self):
        self.document = self.implementation.createDocument(None, 'root', None)
        return self.document

def buildCases(modName, feature, level):
    cases = []
    add = cases.append
    objects = sys.modules[modName].__dict__
    for obj in objects.keys():
        if obj[-8:] != 'TestCase': continue
        add((objects[obj], feature, level))

    return list(cases)



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

from NodePath import BaseNodePathScheme, registry

class ElementIdPathScheme(BaseNodePathScheme):
    def __init__(self, scheme_name=None):
        scheme_name = scheme_name or 'element_id'
        BaseNodePathScheme.__init__(self, scheme_name)

    def resolve_steps(self, top_node, steps):
        node = top_node
        for step in steps:
            if step[0] == 'e':
                element_id = int(step[1:])
                for child in node.childNodes:
                    if self.get_element_id(child) == element_id:
                        node = child
                        break
                else:
                    # couldn't find node with such element_id
                    return None
            else:
                node = node.childNodes.item(int(step))
                if node is None:
                    return None
        return node # return node

    def create_steps(self, top_node, node):
        steps = []
        while node is not top_node:
            parent = node.parentNode
            if parent is None:
                break
            element_id = self.get_element_id(node)
            if element_id != -1:
                steps.append("e%s" % element_id)
            else:
                steps.append(str(parent.childNodes.index(node)))
            node = parent
        steps.reverse()
        return steps

    def get_element_id(self, node):
        return getattr(node, 'elementId', -1)




##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import os, sys
  
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

from Testing import ZopeTestCase

ZopeTestCase.installProduct('ParsedXML')

from Products.ParsedXML import ParsedXML
from domapi import DOMImplementationTestSuite

def ParsedXMLParseString(self, xml):
    return ParsedXML.ParsedXML('foo', xml)

def test_suite():
    """Return a test suite for the Zope testing framework."""
    return DOMImplementationTestSuite(ParsedXML.theDOMImplementation, 
        ParsedXMLParseString)

if __name__ == '__main__':
    framework()



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Replacement DOM exceptions to be used if the xml.dom package is not
available.
"""

"""
Why does this module exist?

    The Python DOM API defines exceptions that DOM implementations
    should use to allow DOM client code to detect errors that can
    occur during processing.  Since not all client code knows about
    the DOM implementation used, all implementations must use shared
    exceptions.  These are defined in the xml.dom package (in the
    package's __init__.py).  The xml.dom package is provided as part
    of PyXML and Python 2.0.

    Since ParsedXML may be used from Python 1.5.2 without having PyXML
    or a more recent version of Python available, we need to provide
    an alternate implementation.  However, DOM client code that works
    on DOM instances created elsewhere will still expect to get the
    exception classes from xml.dom.  Since the code may be part of
    third-party packages that know nothing of ParsedXML or Zope, we
    need to provide an implementation of xml.dom if it doesn't already
    exist.

So how does this module solve the problem?

    This module defines the required exception objects and constants
    and 'installs' the values in the xml.dom module if they are not
    already present.  Since the xml.dom module may not exist, or may
    pre-date the addition of these exceptions to the standard
    implementation (Python 2.0 or PyXML 0.6.2), the modules xml and
    xml.dom are created by surgically adding them to sys.modules if
    needed, and inserting the required values into an existing xml.dom
    module if needed.

    This works because of the way the module import machinery works in
    Python.  sys.modules is a mapping from module name to module
    object; sys.modules['sys'] evaluates to the sys module object.
    When an import statement is executed, the Python runtime first
    looks in sys.modules to retrieve an already-loaded module.  The
    set of built-in modules and the filesystem are only consulted if
    the module has not already been loaded.  For modules in packages
    (xml.dom), each level of enclosing package is checked before
    attempting to load the module; i.e., xml is checked before
    xml.dom.  This machinery is invoked each time an import is
    attempted.

    When ParsedXML.DOM is imported, it imports this module.  This
    first attempts to load the standard xml.dom package.  If that
    fails (which it is likely to do for Python 1.5.2 without PyXML
    installed), this module is an acceptable implementation of
    xml.dom, but we still need the xml package.  This is created
    artificially using the new.module() function and inserted in
    sys.modules.  Once this is done, this module may be inserted for
    the key 'xml.dom', after which attempts to import xml.dom will
    provide this module.

    If xml.dom is already available, but older than the introduction
    of DOMException and its specializations, the implementations
    defined here are inserted into it, so that it is extended to match
    the more recent version of the interface definition.

What are the limitations of this approach?

    Some versions of PyXML may have defined DOMException without
    defining the subclasses.  The specialized versions of DOMException
    were added in PyXML version 0.6.3 (XXX ??).  Versions which
    contain DOMException but not the specializations will not be
    compatible with this module.  This should not be a substantial
    limitation in the context of Zope.

    There is no way to protect against code that imports xml.dom
    before ParsedXML.DOM has been imported.  Such code will receive an
    ImportError.  Reloading that code after ParsedXML.DOM is imported
    will cause it to work properly.

"""

# These have to be in order:
_CODE_NAMES = [
    "INDEX_SIZE_ERR",
    "DOMSTRING_SIZE_ERR",
    "HIERARCHY_REQUEST_ERR",
    "WRONG_DOCUMENT_ERR",
    "INVALID_CHARACTER_ERR",
    "NO_DATA_ALLOWED_ERR",
    "NO_MODIFICATION_ALLOWED_ERR",
    "NOT_FOUND_ERR",
    "NOT_SUPPORTED_ERR",
    "INUSE_ATTRIBUTE_ERR",
    "INVALID_STATE_ERR",
    "SYNTAX_ERR",
    "INVALID_MODIFICATION_ERR",
    "NAMESPACE_ERR",
    "INVALID_ACCESS_ERR",
    ]

for i in range(len(_CODE_NAMES)):
    globals()[_CODE_NAMES[i]] = i + 1
del i


class DOMException(Exception):
    """Base class for exceptions raised by the DOM."""

    def __init__(self, code, *args):
        self.code = code
        self.args = (code,) + args
        Exception.__init__(self, g_errorMessages[code])
        if self.__class__ is DOMException:
            self.__class__ = g_realExceptions[code]

def _derived_init(self, *args):
    """Initializer method that does not expect a code argument,
    for use in derived classes."""
    if not args:
        args = (self, g_errorMessages[self.code])
    else:
        args = (self,) + args
    apply(Exception.__init__, args)

try:
    from xml.dom import DOMException
except ImportError:
    pass

import string
_EXCEPTION_NAMES = ["DOMException"]
template = """\
class %s(DOMException):
    code = %s
    __init__ = _derived_init
"""


g_realExceptions = {}

for s in _CODE_NAMES:
    words = string.split(string.lower(s), "_")
    ename = string.join(map(string.capitalize, words), "")
    exec template % (ename, s)
    g_realExceptions[globals()[s]] = globals()[ename]
    _EXCEPTION_NAMES.append(ename)

del s, words, ename, string, template


try:
    import xml.dom

except ImportError:
    # We have to define everything, which we've done above.
    # This installs it:
    import sys
    try:
        mod = __import__("xml")
    except ImportError:
        import new
        mod = new.module("xml")
        del new
        sys.modules["xml"] = mod
    import Exceptions
    mod.dom = Exceptions
    sys.modules["xml.dom"] = Exceptions
    del mod, sys
    del Exceptions
    from Core import Node

else:
    # The exception classes may not have been defined, so add any
    # that are needed.
    import Exceptions
    for s in _CODE_NAMES + _EXCEPTION_NAMES:
        if not hasattr(xml.dom, s):
            setattr(xml.dom, s, getattr(Exceptions, s))
    if not hasattr(xml.dom, "Node") or type(xml.dom.Node) is type(Exceptions):
        # We need to provide the Node class so the .nodeType constants
        # are in the right place.
        import Core
        xml.dom.Node = Core.Node
        del Core
    del s, Exceptions

del _CODE_NAMES, _EXCEPTION_NAMES


g_errorMessages = {
    INDEX_SIZE_ERR:
    "Index error accessing NodeList or NamedNodeMap",

    DOMSTRING_SIZE_ERR:
    "DOMString exceeds maximum size.",

    HIERARCHY_REQUEST_ERR:
    "Node manipulation results in invalid parent/child relationship.",

    WRONG_DOCUMENT_ERR: "",
    INVALID_CHARACTER_ERR: "",
    NO_DATA_ALLOWED_ERR: "",

    NO_MODIFICATION_ALLOWED_ERR:
    "Attempt to modify a read-only attribute.",

    NOT_FOUND_ERR: "",

    NOT_SUPPORTED_ERR:
    "DOM feature not supported.",

    INUSE_ATTRIBUTE_ERR:
    "Illegal operation on an attribute while in use by an element.",

    INVALID_STATE_ERR: "",
    SYNTAX_ERR: "",
    INVALID_MODIFICATION_ERR: "",

    NAMESPACE_ERR:
    "Namespace operation results in malformed or invalid name or name declaration.",

    INVALID_ACCESS_ERR: "",
    }


# To be sure that unused alternate implementations of the DOM
# exceptions are not used by accessing this module directly, import
# the "right" versions over those defined here.  They may be the same,
# and they may be from an up-to-date PyXML or Python 2.1 or newer.
# This causes alternate implementations to be discarded if not needed.

from xml.dom import *



##############################################################################
#
# Copyright (c) 2001-2004 Zope Corporation and Contributors. All
# Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################

"""Acquisition-based implementation of the DOM 'XML' feature classes."""

import Core
import xml.dom

from ComputedAttribute import ComputedAttribute


class CDATASection(Core.Text):
    nodeName = "#cdata-section"
    nodeType = Core.Node.CDATA_SECTION_NODE


class Identified:
    """Mix-in class that supports the publicId and systemId attributes."""

    def _identified_mixin_init(self, publicId, systemId):
        d = self.__dict__
        d['publicId'] = publicId
        d['systemId'] = systemId

    def _get_publicId(self):
        return self.publicId

    def _get_systemId(self):
        return self.systemId


class Entity(Identified, Core.Parentless, Core.Node):
    nodeType = Core.Node.ENTITY_NODE
    _readonly = 1
    _in_tree = 0

    _allowed_child_types = (Core.Node.ELEMENT_NODE,
                            Core.Node.PROCESSING_INSTRUCTION_NODE,
                            Core.Node.COMMENT_NODE,
                            Core.Node.TEXT_NODE,
                            Core.Node.CDATA_SECTION_NODE,
                            Core.Node.ENTITY_REFERENCE_NODE)

    def __init__(self, name, publicId, systemId, notationName):
        self._identified_mixin_init(publicId, systemId)
        d = self.__dict__
        d['nodeName'] = name
        d['notationName'] = notationName

    def _cloneNode(self, deep, mutable, document):
        # force children to not to acquire mutability:
        return Core.Node._cloneNode(self, deep, 0, document)

    def _get_notationName(self):
        return self.notationName

    # DOM Level 3 (Working Draft, 01 Sep 2000)
    # I expect some or all of these will become read-only before the
    # recommendation is finished.

    actualEncoding = None
    encoding = None
    version = None

    def _get_actualEncoding(self):
        return self.actualEncoding
    def _set_actualEncoding(self, value):
        self.__dict__['actualEncoding'] = value

    def _get_encoding(self):
        return self.encoding
    def _set_encoding(self, value):
        self.__dict__['value'] = value

    def _get_version(self):
        return self.version
    def _set_version(self, value):
        self.__dict__['version'] = value


class EntityReference(Core.Node):
    nodeType = Core.Node.ENTITY_REFERENCE_NODE
    _readonly = 1

    _allowed_child_types = (Core.Node.ELEMENT_NODE,
                            Core.Node.PROCESSING_INSTRUCTION_NODE,
                            Core.Node.COMMENT_NODE,
                            Core.Node.TEXT_NODE,
                            Core.Node.CDATA_SECTION_NODE,
                            Core.Node.ENTITY_REFERENCE_NODE)

    def __init__(self, name):
        self.__dict__['_in_tree'] = 0
        self.__dict__['nodeName'] = name


class Notation(Identified, Core.Childless, Core.Parentless, Core.Node):
    nodeType = Core.Node.NOTATION_NODE
    _readonly = 1

    def __init__(self, name, publicId, systemId):
        self._identified_mixin_init(publicId, systemId)
        d = self.__dict__
        d['_in_tree'] = 0
        d['nodeName'] = name

    def _cloneNode(self, deep, mutable, document):
        # force children to not to acquire mutability:
        return Core.Node._cloneNode(self, deep, 0, document)

    # DOM Level 3 (working draft, 5 June 2001)

    def _get_textContent(self):
        return ''
    textContent = ''


class ProcessingInstruction(Core.Childless, Core.Node):
    nodeType = Core.Node.PROCESSING_INSTRUCTION_NODE

    def __init__(self, target, data):
        d = self.__dict__
        d['_in_tree'] = 0
        d['nodeName'] = target
        d['target'] = target
        d['nodeValue'] = data
        d['data'] = data

    def _get_data(self):
        return self.data

    def _set_data(self, data):
        if self._readonly:
            raise xml.dom.NoModificationAllowedErr()
        d = self.__dict__
        if d['data'] != data:
            d['data'] = data
            d['nodeValue'] = data
            self._changed()

    _set_nodeValue = _set_data

    def _get_target(self):
        return self.target
    target = ComputedAttribute(_get_target, 1)

    # DOM Level 3 (working draft, 5 June 2001)

    def _get_textContent(self):
        return self.nodeValue
    textContent = ComputedAttribute(_get_textContent, 1)


class DocumentType(Identified, Core.Childless, Core.Node):
    nodeType = Core.Node.DOCUMENT_TYPE_NODE
    nodeValue = None
    internalSubset = None

    def __init__(self, qualifiedName, publicId, systemId):
        self._identified_mixin_init(publicId, systemId)
        d = self.__dict__
        d['name'] = qualifiedName
        d['nodeName'] = qualifiedName
        d['_entities'] = []
        d['_notations'] = []
        d['_in_tree'] = 0

    def _get_internalSubset(self):
        return self.internalSubset

    def _get_name(self):
        return self.name

    _get_nodeName = _get_name

    def _set_nodeValue(self, data):
        return

    def _get_entities(self):
        return OwnedEntityMap(self, '_entities')
    entities = ComputedAttribute(_get_entities, 1)

    def _get_notations(self):
        return OwnedEntityMap(self, '_notations')
    notations = ComputedAttribute(_get_notations, 1)

    def isSupported(self, feature, version):
        doc = self.ownerDocument
        if doc:
            impl = doc.implementation
        else:
            impl = Core.theDOMImplementation
        return impl.hasFeature(feature, version)

    # DOM Level 3 (working draft, 5 June 2001)

    def _get_textContent(self):
        return ''
    textContent = ''


class OwnedEntityMap(Core.MapFromParent):
    """
    NamedNodeMap that works on the entity or notation structure
    of a DocumentType.
    """

    def __init__(self, parent, listName):
        Core.MapFromParent.__init__(self, parent)
        self.__dict__['_parentListName'] = listName
        
    def _item_helper(self, itemSource):
        "used by item; create an Attribute from the item and return it"
        # XXX is ownerDocument ok with this?
        #itemSource.__dict__['ownerDocument'] = self._parent
        return itemSource.__of__(self._parent)

    def _nameMatcher(self, itemSource, name):
        return itemSource.nodeName == name

    def _nsMatcher(self, itemSource, namespaceURI, localName):
        return (itemSource.namespaceURI == namespaceURI
                and itemSource.localName == localName)

    def _set_named_item(self, name, matcher, node):
        raise xml.dom.NoModificationAllowedErr()

    def _delFromParentList(self, entities, i):
        raise xml.dom.NoModificationAllowedErr()

    def _addToParentList(self, entities, node):
        raise xml.dom.NoModificationAllowedErr()

    def _key_helper(self, itemSource):
        "Given an item source, return an appropriate key for our mapping"
        if itemSource.prefix:
            return "%s:%s" % (itemSource.prefix,
                              itemSource.localName)
        else:
            return itemSource.localName


# no longer needed
del ComputedAttribute
